'''
Created on Aug 28, 2014

@author: rch
'''

import numpy as np



# Kronecker delta
DELTA = np.zeros((3, 3,), dtype='f')
DELTA[(0, 1, 2), (0, 1, 2)] = 1

# Levi Civita symbol
EPS = np.zeros((3, 3, 3), dtype='f')
EPS[(0, 1, 2), (1, 2, 0), (2, 0, 1)] = 1
EPS[(2, 1, 0), (1, 0, 2), (0, 2, 1)] = -1


def get_sin_theta(a, b):
    r'''Get the cosine of an angle between two vectors :math:`a` and :math:`b`

    .. math::
        \gamma_s =
        \cos{(\theta)}
        =
        \frac{ \dotp{a}{b} }
            { \norm{a} \norm{b} }
        =
        \frac{a_d b_d}
            { \sqrt{a_d a_d} \sqrt{b_d b_d} }
        :label: gamma

    It is convenient to implement this function in a vectorized
    form so that two arrays of vectors can be supplied and
    evaluated using several sum-product evaluations over their
    last index :math:`d` standing for the spatial dimension of
    the vectors.

    Such an evaluation can be conveniently done using the
    method
    `numpy.einsum(subscript_mappings, *operands)
    <http://docs.scipy.org/doc/numpy/reference/generated/numpy.einsum.html>`_.
    Given two compatible arrays (n-dimensional) :math:`a`
    and :math:`b` of vectors with last index :math:`d`
    labeling their spatial component the array :math:`\gamma_c`
    of angle cosines between each pair of vectors in :math:`a`, :math:`b`.
    gets calculated as::

        ab = np.einsum('...d,...d->...', a, b)
        sqrt_aa = np.sqrt(np.einsum('...d,...d->...', a, a))
        sqrt_bb = np.sqrt(np.einsum('...d,...d->...', b, b))
        sqrt_aa_x_sqrt_bb = sqrt_aa * sqrt_bb
        gamma = ab / sqrt_aa_x_sqrt_bb)

    '''
    c = np.einsum('...i,...j,...kij->...k', a, b, EPS)
    mag_c = np.einsum('...i,...i->...', c, c)
    print('c', c, 'mag_c', mag_c)
    mag_aa_bb = np.einsum('...i,...i,...j,...j->...', a, a, b, b)
    print('mag_ab_bb', mag_aa_bb)
    sin_theta = np.sqrt(mag_c / mag_aa_bb)

    return sin_theta


if __name__ == '__main__':

    # end_doc

    a = np.array([[1, 0, 0], [1, 0, 0], [1, 0, 0]], dtype='f')
    b = np.array([[0, 1, 0], [1, 1, 0], [1, -1, 0]], dtype='f')

    print('gamma', get_sin_theta(a, b))



import traits.api as tr
from ibvpy.fets import FETSEval

from bmcs_shell.folding.analysis.fets2d_mitc import FETS2DMITC
from bmcs_shell.folding.analysis.wb_fets2d3u1m_fe import FETS2D3U1M
import bmcs_utils.api as bu
import numpy as np
from ibvpy.mesh.i_fe_uniform_domain import IFEUniformDomain
import k3d

@tr.provides(IFEUniformDomain)
class FETriangularMesh(bu.Model):
    name = 'FETriangularMesh'

    X_Id = tr.Array(np.float_, value=[[0, 0, 0], [2, 0, 0], [2, 2, 0], [1, 1, 0]])
    I_Fi = tr.Array(np.int_, value=[[0, 1, 3],
                                    [1, 2, 3],
                                    ])

    fets = tr.Instance(FETSEval)
    def _fets_default(self):
        return FETS2D3U1M()

    show_node_labels = bu.Bool(False)

    n_nodal_dofs = tr.DelegatesTo('fets')
    dof_offset = tr.Int(0)

    n_active_elems = tr.Property
    def _get_n_active_elems(self):
        return len(self.I_Fi)

    ipw_view = bu.View(
        bu.Item('show_node_labels'),
    )

    #=========================================================================
    # 3d Visualization
    #=========================================================================
    plot_backend = 'k3d'
    show_wireframe = bu.Bool(True)

    def setup_plot(self, pb):
        X_Id = self.X_Id.astype(np.float32)
        I_Fi = self.I_Fi.astype(np.uint32)

        fe_mesh = k3d.mesh(X_Id, I_Fi,
                              color=0x999999, opacity=1.0,
                              side='double')
        pb.plot_fig += fe_mesh
        pb.objects['mesh'] = fe_mesh

        if self.show_wireframe:
            k3d_mesh_wireframe = k3d.mesh(X_Id,
                                          I_Fi,
                                          color=0x000000,
                                          wireframe=True)
            pb.plot_fig += k3d_mesh_wireframe
            pb.objects['mesh_wireframe'] = k3d_mesh_wireframe

        if self.show_node_labels:
            self._add_nodes_labels_to_fig(pb, X_Id)

    NODES_LABELS = 'nodes_labels'

    def update_plot(self, pb):
        X_Id = self.X_Id.astype(np.float32)
        I_Fi = self.I_Fi.astype(np.uint32)

        mesh = pb.objects['mesh']
        mesh.vertices = X_Id
        mesh.indices = I_Fi
        if self.show_wireframe:
            wireframe = pb.objects['mesh_wireframe']
            wireframe.vertices = X_Id
            wireframe.indices = I_Fi

        if self.show_node_labels:
            if self.NODES_LABELS in pb.objects:
                pb.clear_object(self.NODES_LABELS)
            self._add_nodes_labels_to_fig(pb, X_Id)
        else:
            if self.NODES_LABELS in pb.objects:
                pb.clear_object(self.NODES_LABELS)

    def _add_nodes_labels_to_fig(self, pb, X_Id):
        text_list = []
        for I, X_d in enumerate(X_Id):
            k3d_text = k3d.text('%g' % I, tuple(X_d), label_box=False, size=0.8, color=0x00FF00)
            pb.plot_fig += k3d_text
            text_list.append(k3d_text)
        pb.objects[self.NODES_LABELS] = text_list


# TODO Here THERE'S NO TRANSFORMATIONS TO LOCAL WHICH ENABLES COMPARING B MATRIX FROM HERE WITH THE SHELL B MATRIX
#  THE PROBLEM IS IN B MATRIX, WHEN INDICES CHANGE IT'S OK FOR IT TO CHANGE, BUT IT DOESN'T MATCH B FROM THIS PLANE ELEMENT
#  ANYMORE!!

import traits.api as tr
from ibvpy.mathkit.tensor import DELTA23_ab
import numpy as np
from bmcs_shell.folding.utils.vector_acos import \
    get_theta, get_theta_du
import k3d
from ibvpy.mathkit.linalg.sys_mtx_assembly import SysMtxArray

INPUT = '+cp_input'


# Kronecker delta
DELTA = np.zeros((3, 3,), dtype='f')
DELTA[(0, 1, 2), (0, 1, 2)] = 1

# Levi Civita symbol
EPS = np.zeros((3, 3, 3), dtype='f')
EPS[(0, 1, 2), (1, 2, 0), (2, 0, 1)] = 1
EPS[(2, 1, 0), (1, 0, 2), (0, 2, 1)] = -1

from bmcs_shell.folding.analysis.fem.fe_triangular_mesh import FETriangularMesh
from bmcs_shell.folding.analysis.fem.xdomain_fe_grid import XDomainFE

class TriXDomainFE(XDomainFE):
    name = 'TriXDomainFE'
    '''
    Finite element discretization with dofs and mappings derived from the FE definition
    '''

    mesh = tr.Instance(FETriangularMesh, ())
    fets = tr.DelegatesTo('mesh')

    tree = ['mesh']

    change = tr.Event(GEO=True)

    plot_backend = 'k3d'

    n_dofs = tr.Property

    def _get_n_dofs(self):
        return len(self.mesh.X_Id) * self.mesh.n_nodal_dofs

    eta_w = tr.Property
    r'''Weight factors for numerical integration.
    '''

    def _get_eta_w(self):
        return self.fets.w_m

    Na_deta = tr.Property()
    r'''Derivatives of the shape functions in the integration points.
    '''
    @tr.cached_property
    def _get_Na_deta(self):
        return np.einsum('imr->mri', self.fets.dN_imr)

    x_0 = tr.Property()
    r'''Derivatives of the shape functions in the integration points.
    '''
    @tr.cached_property
    def _get_x_0(self):
        return self.mesh.X_Id

    x = tr.Property()
    r'''Derivatives of the shape functions in the integration points.
    '''
    @tr.cached_property
    def _get_x(self):
        return self.x_0

    F = tr.Property()
    r'''Derivatives of the shape functions in the integration points.
    '''
    @tr.cached_property
    def _get_F(self):
        return self.mesh.I_Fi

    T_Fab = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_T_Fab(self):
        return self.F_L_bases[:, 0, :]

    I_Ei = tr.Property
    def _get_I_Ei(self):
        return self.F_N

    x_Eia = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_x_Eia(self):
        X_Eia = self.X_Id[self.I_Ei, :]
        X_E0a = X_Eia[:, 0, :]
        X_Eia -= X_E0a[:, np.newaxis, :]
        # X_Eic = np.einsum('Eac,Eic->Eia', self.T_Fab, X_Eia)
        # return X_Eic[...,:-1]
        return X_Eia[..., :-1]

    def U2u(self, U_Eia):
        u0_Eia = U_Eia[...,:-1]
        return u0_Eia

    def xU2u(self, U_Eia):
        # u1_Eia = np.einsum('Eab,Eib->Eia', self.T_Fab, U_Eia)
        # u2_Eie =  np.einsum('ea,Eia->Eie', DELTA23_ab, u1_Eia)
        u2_Eie = np.einsum('ea,Eia->Eie', DELTA23_ab, U_Eia)
        return u2_Eie

    def f2F(self, f_Eid):
        F0_Eia = np.concatenate( [f_Eid, np.zeros_like(f_Eid[...,:1])], axis=-1)
        return F0_Eia

    def xf2F(self, f_Eid):
        F1_Eia = np.einsum('da,Eid->Eia', DELTA23_ab, f_Eid)
        # F2_Eia = np.einsum('Eab,Eia->Eib', self.T_Fab, F1_Eia)
        # return F2_Eia
        return F1_Eia

    def k2K(self, K_Eiejf):
        K0_Eicjf = np.concatenate( [K_Eiejf, np.zeros_like(K_Eiejf[:,:,:1,:,:])], axis=2)
        K0_Eicjd = np.concatenate( [K0_Eicjf, np.zeros_like(K0_Eicjf[:,:,:,:,:1])], axis=4)
        return K0_Eicjd

    def xk2K(self, K_Eiejf):
        K1_Eiejf = np.einsum('ea,fb,Eiejf->Eiajb', DELTA23_ab, DELTA23_ab, K_Eiejf) # correct
        # T_Eeafb = np.einsum('Eea,Efb->Eeafb', self.T_Fab, self.T_Fab)
        #K_Eab = np.einsum('Eeafb,ef->Eab', T_Eeafb, k_ef)
        # K2_Eiajb = np.einsum('Eeafb,Eiejf->Eiajb', T_Eeafb, K1_Eiejf)
        #K2_Eicjd = np.einsum('Eca,Ebd,Eiajb->Eicjd', self.T_Fab, self.T_Fab, K1_Eicjd)
        #K2_Eicjd = np.einsum('Eac,Edb,Eiajb->Eicjd', self.T_Fab, self.T_Fab, K1_Eicjd)
        return K1_Eiejf

    I_CDij = tr.Property
    def _get_I_CDij(self):
        return self.mesh.I_CDij

    bc_J_F_xyz= tr.Property(depends_on='state_changed')
    @tr.cached_property
    def _get_bc_J_F_xyz(self):
        ix2 = int((self.mesh.n_phi_plus) / 2)
        F_I = self.I_CDij[ix2, :, 0, :].flatten()
        _, idx_remap = self.mesh.unique_node_map
        return idx_remap[F_I]

    bc_J_xyz = tr.Property(depends_on='state_changed')
    @tr.cached_property
    def _get_bc_J_xyz(self):
        I_M = self.I_CDij[(0, -1), :, (0, -1), :].flatten()
        _, idx_remap = self.mesh.unique_node_map
        J_M = idx_remap[I_M]
        return J_M

    bc_J_x = tr.Property(depends_on='state_changed')
    @tr.cached_property
    def _get_bc_J_x(self):
        I_M = self.I_CDij[:, (0, -1), :, (0, -1)].flatten()
        _, idx_remap = self.mesh.unique_node_map
        J_M = idx_remap[I_M]
        return J_M

    def setup_plot(self, pb):
        X_Ia = self.mesh.X_Ia.astype(np.float32)
        I_Fi = self.mesh.I_Fi.astype(np.uint32)

        X_Ma = X_Ia[self.bc_J_xyz]
        self.k3d_fixed_xyz = k3d.points(X_Ma)
        pb.plot_fig += self.k3d_fixed_xyz

        X_Ma = X_Ia[self.bc_J_x]
        self.k3d_fixed_x = k3d.points(X_Ma, color=0x22ffff)
        pb.plot_fig += self.k3d_fixed_x

        X_Ma = X_Ia[self.bc_J_F_xyz]
        self.k3d_load_z = k3d.points(X_Ma, color=0xff22ff)
        pb.plot_fig += self.k3d_load_z

        self.k3d_mesh = k3d.mesh(X_Ia,
                                 I_Fi,
                                 color=0x999999,
                                 side='double')
        pb.plot_fig += self.k3d_mesh

    def update_plot(self, pb):
        X_Ia = self.mesh.X_Ia.astype(np.float32)
        I_Fi = self.mesh.I_Fi.astype(np.uint32)

        self.k3d_fixed_xyz.positions = X_Ia[self.bc_J_xyz]
        self.k3d_fixed_x.positions = X_Ia[self.bc_J_x]
        self.k3d_load_z.positions = X_Ia[self.bc_J_F_xyz]

        mesh = self.k3d_mesh
        mesh.vertices = X_Ia
        mesh.indices = I_Fi

    B_Eso = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_B_Eso(self):
        xx_Ei, yy_Ei = np.einsum('...a->a...', self.X_Id[self.I_Ei, :-1])
        # xx_Ei, yy_Ei = np.einsum('...a->a...', self.mesh.X_Id[self.mesh.I_Fi, :-1])

        y23 = yy_Ei[:, 1] - yy_Ei[:, 2]
        y31 = yy_Ei[:, 2] - yy_Ei[:, 0]
        y12 = yy_Ei[:, 0] - yy_Ei[:, 1]
        x32 = xx_Ei[:, 2] - xx_Ei[:, 1]
        x13 = xx_Ei[:, 0] - xx_Ei[:, 2]
        x21 = xx_Ei[:, 1] - xx_Ei[:, 0]
        x23 = -x32
        y32 = -y23
        y13 = -y31

        J_Ear = np.array([[x13, y13], [x23, y23]])
        J_Ear = np.einsum('ar...->...ar', J_Ear)
        det_J_E = np.linalg.det(J_Ear)

        O = np.zeros_like(y23)
        B_soE = np.array(
            [
                [y23, O, y31, O, y12, O],
                [O, x32, O, x13, O, x21],
                [x32, y23, x13, y31, x21, y12]
            ]
        )
        B_Eso = np.einsum('soE,E->Eso', B_soE, 1 / det_J_E)
        return B_Eso, det_J_E

    def map_U_to_field(self, U_o):
        print('')
        print('U_o,', U_o)
        U_Eia = U_o[self.o_Eia]
        # coordinate transform to local
        u_Eia = self.xU2u(U_Eia)
        u_Eo = u_Eia.reshape(-1, 6)
        B_Eso, _ = self.B_Eso
        eps_Eso = np.einsum(
            'Eso,Eo->Es',
            B_Eso, u_Eo
        )
        print('eps_Eso,', eps_Eso)
        return eps_Eso

    def map_field_to_F(self, sig_Es):
        # print('map_field_to_F:')
        # print('sig_Es:', sig_Es)
        B_Eso, det_J_E = self.B_Eso
        f_Eo = self.integ_factor * np.einsum(
            'Eso,Es,E->Eo',
            B_Eso, sig_Es, det_J_E / 2
        )
        f_Eic = f_Eo.reshape(-1,3,2)
        # coordinate transform to global
        f_Eic = self.xf2F(f_Eic)
        _, n_i, n_c = f_Eic.shape
        f_Ei = f_Eic.reshape(-1, n_i * n_c)
        o_E = self.o_Eia.reshape(-1, n_i * n_c)
        return o_E.flatten(), f_Ei.flatten()

    def map_field_to_K(self, D_Est):
        # print('map_field_to_K:')
        #==========================================================================
        B_Eso, det_J_E = self.B_Eso
        k2_ij = self.integ_factor * np.einsum('Eso,Est,Etp,E->Eop', B_Eso, D_Est, B_Eso, det_J_E / 2)
        K_Eiejf = k2_ij.reshape(-1, 3, 2, 3, 2)
        K_Eicjd = self.xk2K(K_Eiejf)

        _, n_i, n_c, n_j, n_d = K_Eicjd.shape
        K_Eij = K_Eicjd.reshape(-1, n_i * n_c, n_j * n_d)
        o_Ei = self.o_Eia.reshape(-1, n_i * n_c)
        # print('K_Eij:', K_Eij)
        print('o_Ei:', o_Ei)
        return SysMtxArray(mtx_arr=K_Eij, dof_map_arr=o_Ei)


    # =========================================================================
    # Property operators for initial configuration
    # =========================================================================
    F0_normals = tr.Property(tr.Array, depends_on='X, L, F')
    r'''Normal facet vectors.
    '''
    @tr.cached_property
    def _get_F0_normals(self):
        x_F = self.x_0[self.F]
        N_deta_ip = self.Na_deta
        r_deta = np.einsum('ajK,IKi->Iaij', N_deta_ip, x_F)
        Fa_normals = np.einsum('Iai,Iaj,ijk->Iak',
                               r_deta[..., 0], r_deta[..., 1], EPS)
        return np.sum(Fa_normals, axis=1)

    sign_normals = tr.Property(tr.Array, depends_on='X,L,F')
    r'''Orientation of the normal in the initial state.
    This array is used to switch the normal vectors of the faces
    to be oriented in the positive sense of the z-axis.
    '''
    @tr.cached_property
    def _get_sign_normals(self):
        return np.sign(self.F0_normals[:, 2])

    F_N = tr.Property(tr.Array, depends_on='X,L,F')
    r'''Counter-clockwise enumeration.
    '''
    @tr.cached_property
    def _get_F_N(self):
        turn_facets = np.where(self.sign_normals < 0)
        F_N = np.copy(self.F)
        F_N[turn_facets, :] = self.F[turn_facets, ::-1]
        return F_N

    F_normals = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the normals of the facets.
    '''
    @tr.cached_property
    def _get_F_normals(self):
        n = self.Fa_normals
        return np.sum(n, axis=1)

    F_normals_0 = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the normals of the facets.
    '''
    @tr.cached_property
    def _get_F_normals_0(self):
        n = self.Fa_normals_0
        return np.sum(n, axis=1)

    norm_F_normals = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the normed normals of the facets.
    '''
    @tr.cached_property
    def _get_norm_F_normals(self):
        n = self.F_normals
        mag_n = np.sqrt(np.einsum('...i,...i', n, n))
        return n / mag_n[:, np.newaxis]

    norm_F_normals_0 = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the normed normals of the facets.
    '''
    @tr.cached_property
    def _get_norm_F_normals_0(self):
        n = self.F_normals_0
        mag_n = np.sqrt(np.einsum('...i,...i', n, n))
        return n / mag_n[:, np.newaxis]

    # F_normals_du = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Get the normals of the facets.
    # '''
    # @tr.cached_property
    # def _get_F_normals_du(self):
    #     n_du = self.Fa_normals_du
    #     return np.sum(n_du, axis=1)
    #
    # F_area = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Get the surface area of the facets.
    # '''
    # @tr.cached_property
    # def _get_F_area(self):
    #     a = self.Fa_area
    #     A = np.einsum('a,Ia->I', self.eta_w, a)
    #     return A
    #
    # # =========================================================================
    # # Potential energy
    # # =========================================================================
    # F_V = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Get the total potential energy of gravity for each facet
    # '''
    # @tr.cached_property
    # def _get_F_V(self):
    #     eta_w = self.eta_w
    #     a = self.Fa_area
    #     ra = self.Fa_r
    #     F_V = np.einsum('a,Ia,Ia->I', eta_w, ra[..., 2], a)
    #     return F_V
    #
    # F_V_du = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Get the derivative of total potential energy of gravity for each facet
    # with respect to each node and displacement component [FIi]
    # '''
    # @tr.cached_property
    # def _get_F_V_du(self):
    #     r = self.Fa_r
    #     a = self.Fa_area
    #     a_dx = self.Fa_area_du
    #     r3_a_dx = np.einsum('Ia,IaJj->IaJj', r[..., 2], a_dx)
    #     N_eta_ip = self.Na
    #     r3_dx = np.einsum('aK,KJ,j->aJj', N_eta_ip, DELTA, DELTA[2, :])
    #     a_r3_dx = np.einsum('Ia,aJj->IaJj', a, r3_dx)
    #     F_V_du = np.einsum('a,IaJj->IJj', self.eta_w, (a_r3_dx + r3_a_dx))
    #     return F_V_du
    #
    # =========================================================================
    # Line vectors
    # =========================================================================
    #
    # F_L_vectors_0 = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Get the cycled line vectors around the facet
    # The cycle is closed - the first and last vector are identical.
    #
    # .. math::
    #     v_{pld} \;\mathrm{where} \; p\in\mathcal{F}, l\in (0,1,2), d\in (0,1,2)
    #
    # with the indices :math:`p,l,d` representing the facet, line vector around
    # the facet and and vector component, respectively.
    # '''
    # @tr.cached_property
    # def _get_F_L_vectors_0(self):
    #     F_N = self.F_N  # F_N is cycled counter clockwise
    #     return self.x_0[F_N[:, (1, 2, 0)]] - self.x_0[F_N[:, (0, 1, 2)]]
    #
    # F_L_vectors = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Get the cycled line vectors around the facet
    # The cycle is closed - the first and last vector are identical.
    #
    # .. math::
    #     v_{pld} \;\mathrm{where} \; p\in\mathcal{F}, l\in (0,1,2), d\in (0,1,2)
    #
    # with the indices :math:`p,l,d` representing the facet, line vector around
    # the facet and and vector component, respectively.
    # '''
    # @tr.cached_property
    # def _get_F_L_vectors(self):
    #     F_N = self.F_N  # F_N is cycled counter clockwise
    #     return self.x[F_N[:, (1, 2, 0)]] - self.x[F_N[:, (0, 1, 2)]]
    #
    # F_L_vectors_du = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Get the derivatives of the line vectors around the facets.
    #
    # .. math::
    #     \pard{v_{pld}}{x_{Ie}} \; \mathrm{where} \;
    #     p \in \mathcal{F},  \in (0,1,2), d\in (0,1,2), I\in \mathcal{N},
    #     e \in (0,1,3)
    #
    # with the indices :math:`p,l,d,I,e` representing the facet,
    # line vector around the facet and and vector component,
    # node vector and and its component index,
    # respectively.
    #
    # This array works essentially as an index function delivering -1
    # for the components of the first node in each dimension and +1
    # for the components of the second node
    # in each dimension.
    #
    # For a facet :math:`p` with lines :math:`l` and component :math:`d` return
    # the derivatives with respect to the displacement of the node :math:`I`
    # in the direction :math:`e`.
    #
    # .. math::
    #     \bm{a}_1 = \bm{x}_2 - \bm{x}_1 \\
    #     \bm{a}_2 = \bm{x}_3 - \bm{x}_2 \\
    #     \bm{a}_3 = \bm{x}_1 - \bm{x}_3
    #
    # The corresponding derivatives are then
    #
    # .. math::
    #     \pard{\bm{a}_1}{\bm{u}_1} = -1, \;\;\;
    #     \pard{\bm{a}_1}{\bm{u}_2} = 1 \\
    #     \pard{\bm{a}_2}{\bm{u}_2} = -1, \;\;\;
    #     \pard{\bm{a}_2}{\bm{u}_3} = 1 \\
    #     \pard{\bm{a}_3}{\bm{u}_3} = -1, \;\;\;
    #     \pard{\bm{a}_3}{\bm{u}_1} = 1 \\
    #
    # '''
    #
    # def _get_F_L_vectors_du(self):
    #     return self.L_vectors_du[self.F_L]
    #
    # F_L_vectors_dul = tr.Property(tr.Array, depends_on=INPUT)
    #
    # def _get_F_L_vectors_dul(self):
    #     return self.L_vectors_dul[self.F_L]
    #
    # norm_F_L_vectors = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Get the cycled line vectors around the facet
    # The cycle is closed - the first and last vector are identical.
    # '''
    # @tr.cached_property
    # def _get_norm_F_L_vectors(self):
    #     v = self.F_L_vectors
    #     mag_v = np.sqrt(np.einsum('...i,...i', v, v))
    #     return v / mag_v[..., np.newaxis]
    #
    # norm_F_L_vectors_du = tr.Property(tr.Array, depends_on=INPUT)
    # '''Get the derivatives of cycled line vectors around the facet
    # '''
    # @tr.cached_property
    # def _get_norm_F_L_vectors_du(self):
    #     v = self.F_L_vectors
    #     v_du = self.F_L_vectors_du  # @UnusedVariable
    #     mag_v = np.einsum('...i,...i', v, v)  # @UnusedVariable
    #     # @todo: finish the chain rule
    #     raise NotImplemented
    #
    # # =========================================================================
    # # Orthonormal basis of each facet.
    # # =========================================================================
    # F_L_bases = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Line bases around a facet.
    # '''
    # @tr.cached_property
    # def _get_F_L_bases(self):
    #     l = self.norm_F_L_vectors
    #     n = self.norm_F_normals
    #     lxn = np.einsum('...li,...j,...kij->...lk', l, n, EPS)
    #     n_ = n[:, np.newaxis, :] * np.ones((1, 3, 1), dtype='float_')
    #     T = np.concatenate([l[:, :, np.newaxis, :],
    #                         -lxn[:, :, np.newaxis, :],
    #                         n_[:, :, np.newaxis, :]], axis=2)
    #     return T
    #
    # F_L_bases_du = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Derivatives of the line bases around a facet.
    # '''
    # @tr.cached_property
    # def _get_F_L_bases_du(self):
    #     '''Derivatives of line bases'''
    #     raise NotImplemented
    #
    # # =========================================================================
    # # Sector angles
    # # =========================================================================
    # F_theta = tr.Property(tr.Array, depends_on=INPUT)
    # '''Get the sector angles :math:`\theta`  within a facet.
    # '''
    # @tr.cached_property
    # def _get_F_theta(self):
    #     v = self.F_L_vectors
    #     a = -v[:, (2, 0, 1), :]
    #     b = v[:, (0, 1, 2), :]
    #     return get_theta(a, b)
    #
    # F_theta_du = tr.Property(tr.Array, depends_on=INPUT)
    # r'''Get the derivatives of sector angles :math:`\theta` within a facet.
    # '''
    # @tr.cached_property
    # def _get_F_theta_du(self):
    #     v = self.F_L_vectors
    #     v_du = self.F_L_vectors_du
    #
    #     a = -v[:, (2, 0, 1), :]
    #     b = v[:, (0, 1, 2), :]
    #     a_du = -v_du[:, (2, 0, 1), ...]
    #     b_du = v_du[:, (0, 1, 2), ...]
    #
    #     return get_theta_du(a, a_du, b, b_du)
    #
    # Fa_normals_du = tr.Property
    # '''Get the derivatives of the normals with respect
    # to the node displacements.
    # '''
    #
    # def _get_Fa_normals_du(self):
    #     x_F = self.x[self.F_N]
    #     N_deta_ip = self.Na_deta
    #     NN_delta_eps_x1 = np.einsum('aK,aL,KJ,dli,ILl->IaiJd',
    #                                 N_deta_ip[:, 0, :], N_deta_ip[:, 1, :],
    #                                 DELTA, EPS, x_F)
    #     NN_delta_eps_x2 = np.einsum('aK,aL,LJ,kdi,IKk->IaiJd',
    #                                 N_deta_ip[:, 0, :], N_deta_ip[:, 1, :],
    #                                 DELTA, EPS, x_F)
    #     n_du = NN_delta_eps_x1 + NN_delta_eps_x2
    #     return n_du
    #
    # Fa_area_du = tr.Property
    # '''Get the derivatives of the facet area with respect
    # to node displacements.
    # '''
    #
    # def _get_Fa_area_du(self):
    #     a = self.Fa_area
    #     n = self.Fa_normals
    #     n_du = self.Fa_normals_du
    #     a_du = np.einsum('Ia,Iak,IakJd->IaJd', 1 / a, n, n_du)
    #     return a_du
    #
    # Fa_normals = tr.Property
    # '''Get normals of the facets.
    # '''
    #
    # def _get_Fa_normals(self):
    #     x_F = self.x[self.F_N]
    #     N_deta_ip = self.Na_deta
    #     r_deta = np.einsum('ajK,IKi->Iaij', N_deta_ip, x_F)
    #     return np.einsum('Iai,Iaj,ijk->Iak',
    #                      r_deta[..., 0], r_deta[..., 1], EPS)
    #
    # Fa_normals_0 = tr.Property
    # '''Get normals of the facets.
    # '''
    #
    # def _get_Fa_normals_0(self):
    #     x_F = self.x_0[self.F_N]
    #     N_deta_ip = self.Na_deta
    #     r_deta = np.einsum('ajK,IKi->Iaij', N_deta_ip, x_F)
    #     return np.einsum('Iai,Iaj,ijk->Iak',
    #                      r_deta[..., 0], r_deta[..., 1], EPS)
    #
    # Fa_area = tr.Property
    # '''Get the surface area of the facets.
    # '''
    #
    # def _get_Fa_area(self):
    #     n = self.Fa_normals
    #     a = np.sqrt(np.einsum('Iai,Iai->Ia', n, n))
    #     return a
    #
    # Fa_r = tr.Property
    # '''Get the reference vector to integrations point in each facet.
    # '''
    #
    # def _get_Fa_r(self):
    #     x_F = self.x[self.F_N]
    #     N_eta_ip = self.Na
    #     r = np.einsum('aK,IKi->Iai', N_eta_ip, x_F)
    #     return r

    # =========================================================================
    # Interier as level set
    # =========================================================================



'''
Created on Feb 8, 2018

@author: rch
'''

from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
import numpy as np


class MATSShellElastic(MATS2DEval):
    '''Elastic material model.
    '''
    state_var_shapes = {}

    def get_corr_pred(self, eps_Es, t_n1):

        E_ = self.E # 1000# 70e+3
        nu_ = self.nu # 0 # 0.3
        # k: Shear correction factor (See P. 474 FEM by Zienkiewicz (ISBN: 1856176339))
        k = 5/6
        D_st_ = E_ / (1 - nu_ ** 2) * np.array([[1, nu_, 0, 0, 0],
                                                [nu_, 1, 0, 0, 0],
                                                [0, 0, 0.5 * (1 - nu_),     0, 0],
                                                [0, 0, 0, 0.5 * (1 - nu_) * k, 0],
                                                [0, 0, 0, 0, 0.5 * (1 - nu_) * k]])
        D_Est = D_st_[np.newaxis, :, :]
        sig_Es = np.einsum('st,...t->...s', D_st_, eps_Es)

        return sig_Es, D_Est



import random

import k3d
import numpy as np
import traits.api as tr
from scipy.optimize import minimize

from bmcs_shell.folding.geometry.wb_tessellation.wb_tessellation_base import WBTessellationBase


class WBNumTessellationBase(WBTessellationBase):
    name = 'WB Num. Tessellation Base'

    sol = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_sol(self):
        print('---------------------------')
        sol = self.minimize_dist()
        # Transfer angles to range [-pi, pi] (to avoid having angle > 2pi so we can do the comparison that follows)
        sol = np.arctan2(np.sin(sol), np.cos(sol))
        print('num_sol=', sol)
        return sol

    def minimize_dist(self):
        x0 = np.array([np.pi, np.pi])
        try:
            res = minimize(self.rotate_and_get_diff, x0, tol=1e-4)
        except:
            print('Error while minimizing!')
            return np.array([0, 0])
        smallest_dist = res.fun
        print('smallest_dist=', smallest_dist)
        sol = res.x
        return sol

    def rotate_and_get_diff(self, rotations):
        br_X_Ia_rot = self._get_br_X_Ia(self.X_Ia, rot=rotations[0])
        ur_X_Ia_rot = self._get_ur_X_Ia(self.X_Ia, rot=rotations[1])
        diff = ur_X_Ia_rot[1] - br_X_Ia_rot[3]
        dist = np.sqrt(np.sum(diff * diff))
        #     print('dist=', dist)
        return dist



import bmcs_utils.api as bu
import k3d

from bmcs_shell.folding.geometry.wb_tessellation.wb_num_tessellation_base import WBNumTessellationBase
import numpy as np

class WBNumTessellation(WBNumTessellationBase):
    name = 'WBNumTessellation'

    n_y = bu.Int(3, GEO=True)
    n_x = bu.Int(3, GEO=True)

    ipw_view = bu.View(
        *WBNumTessellationBase.ipw_view.content,
        bu.Item('n_x', latex=r'n_x'),
        bu.Item('n_y', latex=r'n_y'),
    )

    def calc_mesh_for_tessellated_cells(self):
        # TODO: the resulting mesh_X_Ia, mesh_I_Fi are just summing up all cells, repeation deletion is needed to use
        #  it in analysis
        I_Fi = self.I_Fi
        X_Ia = self.X_Ia

        y_base_cell_X_Ia = X_Ia
        next_y_base_cell_X_Ia = X_Ia
        base_cell_X_Ia = X_Ia
        next_base_cell_X_Ia = X_Ia

        n_y, n_x = self.n_y, self.n_x

        mesh_X_Ia = np.zeros((n_y, n_x, 7, 3))

        # Calculating mesh_I_Fi
        seven_mult_i = 7 * np.arange(n_x * n_y)
        seven_mult_Iab = seven_mult_i[:, np.newaxis, np.newaxis] + np.zeros((n_x * n_y, 6, 3), dtype=np.int32)
        mesh_I_Fi = np.full((n_x * n_y, 6, 3), I_Fi)
        mesh_I_Fi = mesh_I_Fi + seven_mult_Iab
        mesh_I_Fi.reshape((n_x * n_y * 6, 3))

        for i in range(n_y):
            i_row_is_even = (i + 1) % 2 == 0

            add_br = True  # to switch between adding br and ur
            add_bl = True  # to switch between adding bl and ul

            for j in range(n_x):
                if j == 0:
                    mesh_X_Ia[i, j, ...] = base_cell_X_Ia
                    # self.add_cell_to_pb(pb, base_cell_X_Ia, I_Fi, '')
                    continue
                if (j + 1) % 2 == 0:
                    # Number of cell_to_add is even (add right from base cell)
                    if add_br:
                        cell_to_add = self._get_br_X_Ia(base_cell_X_Ia)
                        mesh_X_Ia[i, j, ...] = cell_to_add
                        # self.add_cell_to_pb(pb, cell_to_add, I_Fi, '')
                    else:
                        cell_to_add = self._get_ur_X_Ia(base_cell_X_Ia)
                        mesh_X_Ia[i, j, ...] = cell_to_add
                        # self.add_cell_to_pb(pb, cell_to_add, I_Fi, '')
                    add_br = not add_br
                    base_cell_X_Ia = next_base_cell_X_Ia
                    next_base_cell_X_Ia = cell_to_add
                else:
                    # Number of cell_to_add is odd (add left from base cell)
                    if add_bl:
                        cell_to_add = self._get_bl_X_Ia(base_cell_X_Ia)
                        mesh_X_Ia[i, j, ...] = cell_to_add
                        # self.add_cell_to_pb(pb, cell_to_add, I_Fi, '')
                    else:
                        cell_to_add = self._get_ul_X_Ia(base_cell_X_Ia)
                        mesh_X_Ia[i, j, ...] = cell_to_add
                        # self.add_cell_to_pb(pb, cell_to_add, I_Fi, '')
                    add_bl = not add_bl
                    base_cell_X_Ia = next_base_cell_X_Ia
                    next_base_cell_X_Ia = cell_to_add

            if i_row_is_even:
                # Next row is odd (change y_base_cell_X_Ia to a cell below base cell)
                base_cell_X_Ia = self._get_bl_X_Ia(self._get_br_X_Ia(next_y_base_cell_X_Ia))
                next_base_cell_X_Ia = base_cell_X_Ia
                next_y_base_cell_X_Ia = y_base_cell_X_Ia
                y_base_cell_X_Ia = base_cell_X_Ia
            else:
                # Next row is even (change y_base_cell_X_Ia to a cell above base cell)
                base_cell_X_Ia = self._get_ul_X_Ia(self._get_ur_X_Ia(next_y_base_cell_X_Ia))
                next_base_cell_X_Ia = base_cell_X_Ia
                next_y_base_cell_X_Ia = y_base_cell_X_Ia
                y_base_cell_X_Ia = base_cell_X_Ia

        mesh_X_Ia.reshape((n_x * n_y * 7, 3))
        return mesh_X_Ia, mesh_I_Fi

    # Plotting ##########################################################################

    def setup_plot(self, pb):
        self.pb = pb
        pb.clear_fig()
        X_Ia, I_Fi = self.calc_mesh_for_tessellated_cells()
        self.add_cell_to_pb(pb, X_Ia, I_Fi, 'wb_tess_mesh')

    def update_plot(self, pb):
        if self.k3d_mesh:
            X_Ia, I_Fi  = self.calc_mesh_for_tessellated_cells()
            X_Ia = X_Ia.astype(np.float32)
            I_Fi = I_Fi.astype(np.uint32)
            self.k3d_mesh['wb_tess_mesh'].vertices = X_Ia
            self.k3d_mesh['wb_tess_mesh'].indices = I_Fi
            self.k3d_wireframe['wb_tess_mesh'].vertices = X_Ia
            self.k3d_wireframe['wb_tess_mesh'].indices = I_Fi
        else:
            self.setup_plot(pb)



import numpy as np

# Source:
# https://github.com/nghiaho12/rigid_transform_3D
# http://nghiaho.com/?page_id=671
# Input: expects 3xN matrix of points (A and B are two sets of points)
# Returns R,t (best rotation and translation that transform from set A to B)
# R = 3x3 rotation matrix
# t = 3x1 column vector
def get_best_rot_and_trans_3d(A, B):
    assert A.shape == B.shape

    num_rows, num_cols = A.shape
    if num_rows != 3:
        raise Exception(f"matrix A is not 3xN, it is {num_rows}x{num_cols}")

    num_rows, num_cols = B.shape
    if num_rows != 3:
        raise Exception(f"matrix B is not 3xN, it is {num_rows}x{num_cols}")

    # find mean column wise
    centroid_A = np.mean(A, axis=1)
    centroid_B = np.mean(B, axis=1)

    # ensure centroids are 3x1
    centroid_A = centroid_A.reshape(-1, 1)
    centroid_B = centroid_B.reshape(-1, 1)

    # subtract mean
    Am = A - centroid_A
    Bm = B - centroid_B

    H = Am @ np.transpose(Bm)

    # sanity check
    # if linalg.matrix_rank(H) < 3:
    #    raise ValueError("rank of H = {}, expecting 3".format(linalg.matrix_rank(H)))

    # find rotation
    U, S, Vt = np.linalg.svd(H)
    R = Vt.T @ U.T

    # special reflection case
    if np.linalg.det(R) < 0:
        # print("det(R) < R, reflection detected!, correcting for it ...")
        Vt[2, :] *= -1
        R = Vt.T @ U.T

    t = -R @ centroid_A + centroid_B

    return R, t


def get_rot_matrix_around_vector(v, angle):
    c = np.cos(angle)
    s = np.sin(angle)
    v_norm = v / np.sqrt(sum(v * v))

    # See: Rotation matrix from axis and angle (https://en.wikipedia.org/wiki/Rotation_matrix)
    cross_product_matrix = np.cross(v_norm, np.identity(v_norm.shape[0]) * -1)
    return c * np.identity(3) + s * cross_product_matrix + (1 - c) * np.outer(v_norm, v_norm)


"""

"""
import bmcs_utils.api as bu
import sympy as sp
from bmcs_shell.folding.geometry.wb_cell.wb_cell import WBCell
from sympy.algebras.quaternion import Quaternion
import k3d
import traits.api as tr
import numpy as np


class WBCellSymb4Param(bu.SymbExpr):

    a, b, c = sp.symbols('a, b, c', positive=True)
    u_2, u_3 = sp.symbols('u_2, u_3', positive=True)
    gamma = sp.symbols('gamma', positive=True)

    U0_a = sp.Matrix([a, b, 0])
    W0_a = sp.Matrix([c, 0, 0])
    UW0_a = W0_a - U0_a
    L2_U_0 = (U0_a.T * U0_a)[0]
    L2_UW_0 = (UW0_a.T * UW0_a)[0]

    U1_a = sp.Matrix([a, u_2, u_3])
    W1_a = sp.Matrix([c * sp.sin(gamma), 0, c * sp.cos(gamma)])
    UW1_a = U1_a - W1_a
    L2_U_1 = (U1_a.T * U1_a)[0]
    L2_UW_1 = (UW1_a.T * UW1_a)[0]

    u2_sol = sp.solve(L2_U_1 - L2_U_0, u_2)[0]
    u3_sol = sp.solve((L2_UW_1 - L2_UW_0).subs(u_2, u2_sol), u_3)[0]
    u_3_ = u3_sol
    u_2_ = u2_sol.subs(u_3, u3_sol)

    U_pp_a = U1_a.subs({u_2: u_2_, u_3: u_3_})
    U_mm_a = sp.Matrix([-U_pp_a[0], -U_pp_a[1], U_pp_a[2]])
    U_mp_a = sp.Matrix([-U_pp_a[0], U_pp_a[1], U_pp_a[2]])
    U_pm_a = sp.Matrix([U_pp_a[0], -U_pp_a[1], U_pp_a[2]])
    W_p_a = W1_a.subs({u_2: u_2_, u_3: u_3_})
    W_m_a = sp.Matrix([-W_p_a[0], W_p_a[1], W_p_a[2]])

    V_UW = U_pp_a - W_p_a
    L_UW = sp.sqrt(V_UW[1] ** 2 + V_UW[2] ** 2)
    theta_sol = sp.simplify(2 * sp.asin(V_UW[2] / L_UW))

    theta = sp.Symbol(r'theta')
    q_theta = Quaternion.from_axis_angle([1, 0, 0], theta)

    d_1, d_2, d_3 = sp.symbols('d_1, d_2, d_3')
    D_a = sp.Matrix([d_1, d_2, d_3])
    UD_pp_a = U_pp_a + D_a
    WD_p_a = W_p_a + D_a
    d_subs = sp.solve(UD_pp_a - W_m_a, [d_1, d_2, d_3])

    # center of rotation
    UD_pp_a_ = UD_pp_a.subs(d_subs)
    # rotated point
    WD_p_a_ = WD_p_a.subs(d_subs)
    # pull back
    WD_p_a_pb = WD_p_a_ - UD_pp_a_
    # rotate using quaternion
    WD_p_a_rot = q_theta.rotate_point(WD_p_a_pb.T, q_theta)
    # push forward
    WD_p_a_pf = sp.Matrix(WD_p_a_rot) + UD_pp_a_
    # rotated compatibility point
    WD_p_a_theta = WD_p_a_pf.subs(theta, -theta_sol)

    # rotate the center of the neighbour cell
    DD_a_pb = D_a.subs(d_subs) - UD_pp_a_
    DD_a_rot = q_theta.rotate_point(DD_a_pb.T, q_theta)
    DD_a_pf = sp.simplify(sp.Matrix(DD_a_rot) + UD_pp_a_)
    DD_a_theta = DD_a_pf.subs(theta, -theta_sol)

    H = W_p_a[2]

    rho = (U_mm_a[2] - DD_a_theta[2]) / (U_mm_a[1] - DD_a_theta[1]) * U_mm_a[1]
    R_0 = U_mm_a[2] - rho
    delta_phi = sp.asin(DD_a_theta[1] / R_0)
    delta_x = a + W_p_a[0]

    # theta = sp.symbols('theta')
    # x_1, x_2, x_3 = sp.symbols('x_1, x_2, x_3')
    #
    # q_theta = Quaternion.from_axis_angle([1, 0, 0], theta)
    # X_rot = q_theta.rotate_point((x_1, x_2, x_3), q_theta)
    # X_theta_a = sp.simplify(sp.Matrix(X_rot))

    symb_model_params = ['gamma', 'a', 'b', 'c', ]
    symb_expressions = [
        ('u_2_', ()),
        ('u_3_', ()),
        ('R_0', ()),
        ('delta_phi', ()),
        ('delta_x', ()),
        ('H', ()),
        ('theta_sol', ())
    ]


class WBCell4Param(WBCell, bu.InjectSymbExpr):
    name = 'Waterbomb cell 4p'
    symb_class = WBCellSymb4Param

    plot_backend = 'k3d'

    gamma = bu.Float(np.pi/2-0.001, GEO=True)
    a = bu.Float(1000, GEO=True)
    b = bu.Float(1000, GEO=True)
    c = bu.Float(1000, GEO=True)
    a_high = bu.Float(2000)
    b_high = bu.Float(2000)
    c_high = bu.Float(2000)

    ipw_view = bu.View(
        bu.Item('gamma', latex=r'\gamma', editor=bu.FloatRangeEditor(
            low=1e-6, high=np.pi / 2, n_steps=100, continuous_update=True)),
        bu.Item('a', latex='a', editor=bu.FloatRangeEditor(
            low=1e-6, high_name='a_high', n_steps=100, continuous_update=True)),
        bu.Item('b', latex='b', editor=bu.FloatRangeEditor(
            low=1e-6, high_name='b_high', n_steps=100, continuous_update=True)),
        bu.Item('c', latex='c', editor=bu.FloatRangeEditor(
            low=1e-6, high_name='c_high', n_steps=100, continuous_update=True)),
        *WBCell.ipw_view.content,
    )

    n_I = tr.Property
    def _get_n_I(self):
        return len(self.X_Ia)

    X_Ia = tr.Property(depends_on='+GEO')
    '''Array with nodal coordinates I - node, a - dimension
    '''

    @tr.cached_property
    def _get_X_Ia(self):
        gamma = self.gamma
        u_2 = self.symb.get_u_2_()
        u_3 = self.symb.get_u_3_()
        return np.array([
            [0, 0, 0],  # 0 point
            [self.a, u_2, u_3],  # U++
            [-self.a, u_2, u_3],  # U-+
            [self.a, -u_2, u_3],  # U+-
            [-self.a, -u_2, u_3],  # U--
            [self.c * np.sin(gamma), 0, self.c * np.cos(gamma)],  # W0+
            [-self.c * np.sin(gamma), 0, self.c * np.cos(gamma)]  # W0-
        ], dtype=np.float_
        )

    I_boundary = tr.Array(np.int_, value=[[2, 1],
                                          [6, 5],
                                          [4, 3], ])
    '''Boundary nodes in 2D array to allow for generation of shell boundary nodes'''

    # X_theta_Ia = tr.Property(depends_on='+GEO')
    # '''Array with nodal coordinates I - node, a - dimension
    # '''
    # @tr.cached_property
    # def _get_X_theta_Ia(self):
    #     D_a = self.symb.get_D_(self.gamma).T
    #     theta = self.symb.get_theta_sol(self.gamma)
    #     XD_Ia = D_a + self.X_Ia
    #     X_center = XD_Ia[1,:]
    #
    #     rotation_axes = np.array([[1, 0, 0]], dtype=np.float_)
    #     rotation_angles = np.array([-theta], dtype=np.float_)
    #     rotation_centers = np.array([X_center], dtype=np.float_)
    #
    #     x_single = np.array([XD_Ia], dtype='f')
    #     x_pulled_back = x_single - rotation_centers[:, np.newaxis, :]
    #     q = axis_angle_to_q(rotation_axes, rotation_angles)
    #     x_rotated = qv_mult(q, x_pulled_back)
    #     x_pushed_forward = x_rotated + rotation_centers[:, np.newaxis, :]
    #     x_translated = x_pushed_forward #  + self.translations[:, np.newaxis, :]
    #     return x_translated[0,...]

    delta_x = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_delta_x(self):
        return self.symb.get_delta_x()

    delta_phi = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_delta_phi(self):
        return self.symb.get_delta_phi()

    R_0 = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_R_0(self):
        return self.symb.get_R_0()


def q_normalize(q, axis=1):
    sq = np.sqrt(np.sum(q * q, axis=axis))
    sq[np.where(sq == 0)] = 1.e-19
    return q / sq[:, np.newaxis]


def v_normalize(q, axis=1):
    sq = np.einsum('...a,...a->...', q, q)
    sq[np.where(sq == 0)] = 1.e-19
    return q / sq[..., np.newaxis]


def q_mult(q1, q2):
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2
    w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
    x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
    y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
    z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
    return np.array([w, x, y, z], dtype='f')


def q_conjugate(q):
    qn = q_normalize(q.T).T
    w, x, y, z = qn
    return np.array([w, -x, -y, -z], dtype='f')


def qv_mult(q1, u):
    #print('q1', q1.shape, 'u', u.shape)
    zero_re = np.zeros((u.shape[0], u.shape[1]), dtype='f')
    #print('zero_re', zero_re.shape)
    q2 = np.concatenate([zero_re[:, :, np.newaxis], u], axis=2)
    #print('q2', q2.shape)
    q2 = np.rollaxis(q2, 2)
    #print('q2', q2.shape)
    q12 = q_mult(q1[:, :, np.newaxis], q2[:, :, :])
    #print('q12', q12.shape)
    q_con = q_conjugate(q1)
    #print('q_con', q_con.shape)
    q = q_mult(q12, q_con[:, :, np.newaxis])
    #print('q', q.shape)
    q = np.rollaxis(np.rollaxis(q, 2), 2)
    #print('q', q.shape)
    return q[:, :, 1:]


def axis_angle_to_q(v, theta):
    v_ = v_normalize(v, axis=1)
    x, y, z = v_.T
#    print('x,y,z', x, y, z)
    theta = theta / 2
#    print('theta', theta)
    w = np.cos(theta)
    x = x * np.sin(theta)
    y = y * np.sin(theta)
    z = z * np.sin(theta)
#    print('x,y,z', x, y, z)
    return np.array([w, x, y, z], dtype='f')


def q_to_axis_angle(q):
    w, v = q[0, :], q[1:, :]
    theta = np.arccos(w) * 2.0
    return theta, v_normalize(v)




import errno
import os
import subprocess
import sys
import time


PIPE = subprocess.PIPE

if subprocess._mswindows:
    from win32file import ReadFile, WriteFile
    from win32pipe import PeekNamedPipe
    import msvcrt
else:
    import select
    import fcntl

 
class Popen(subprocess.Popen):
    '''
    This class is a code-snipped from: 
    http://www.google.de/url?sa=t&rct=j&q=&esrc=s&source=web&cd=1&ved=0CDgQFjAA&url=http%3A%2F%2Fcode.activestate.com%2Frecipes%2F440554-module-to-allow-asynchronous-subprocess-use-on-win%2Fdownload%2F1%2F&ei=L81SUc6YFdOM4gSWoYHoBQ&usg=AFQjCNEpnjK5yQO8OdMuevJP4gToOp7yvQ&sig2=DgbTCRxkDcR3PSbBOWFdhA&bvm=bv.44342787,d.bGE&cad=rja
    
    ToDo:
    This code can't interpret answers from the cluster,
    so it just waits some time (default 5 sec) and then 
    starting the next input. So it won't catch any errors!!!
    '''
    def recv(self, maxsize = None):
        '''
        Standard output stream from cluster.
        '''
        return self._recv('stdout', maxsize)

    def recv_err(self, maxsize = None):
        '''
        Standard error stream from cluster.
        '''
        return self._recv('stderr', maxsize)

    def send_recv(self, input = '', maxsize = None):
        return self.send(input), self.recv(maxsize), self.recv_err(maxsize)

    def get_conn_maxsize(self, which, maxsize):
        if maxsize is None:
            maxsize = 1024
        elif maxsize < 1:
            maxsize = 1
        return getattr(self, which), maxsize

    def _close(self, which):
        getattr(self, which).close()
        setattr(self, which, None)

    if subprocess._mswindows:
        def send(self, input):
            if not self.stdin:
                return None

            try:
                x = msvcrt.get_osfhandle(self.stdin.fileno())
                (errCode, written) = WriteFile(x, input)
            except ValueError:
                return self._close('stdin')
            except (subprocess.pywintypes.error, Exception) as why:
                if why[0] in (109, errno.ESHUTDOWN):
                    return self._close('stdin')
                raise

            return written

        def _recv(self, which, maxsize):
            conn, maxsize = self.get_conn_maxsize(which, maxsize)
            if conn is None:
                return None

            try:
                x = msvcrt.get_osfhandle(conn.fileno())
                (read, nAvail, nMessage) = PeekNamedPipe(x, 0)
                if maxsize < nAvail:
                    nAvail = maxsize
                if nAvail > 0:
                    (errCode, read) = ReadFile(x, nAvail, None)
            except ValueError:
                return self._close(which)
            except (subprocess.pywintypes.error, Exception) as why:
                if why[0] in (109, errno.ESHUTDOWN):
                    return self._close(which)
                raise

            if self.universal_newlines:
                read = self._translate_newlines(read)
            return read

    else:
        def send(self, input):
            if not self.stdin:
                return None

            if not select.select([], [self.stdin], [], 0)[1]:
                return 0

            try:
                written = os.write(self.stdin.fileno(), input)
            except OSError as why:
                if why[0] == errno.EPIPE: #broken pipe
                    return self._close('stdin')
                raise

            return written

        def _recv(self, which, maxsize):
            conn, maxsize = self.get_conn_maxsize(which, maxsize)
            if conn is None:
                return None

            flags = fcntl.fcntl(conn, fcntl.F_GETFL)
            if not conn.closed:
                fcntl.fcntl(conn, fcntl.F_SETFL, flags | os.O_NONBLOCK)

            try:
                if not select.select([conn], [], [], 0)[0]:
                    return ''

                r = conn.read(maxsize)
                if not r:
                    return self._close(which)

                if self.universal_newlines:
                    r = self._translate_newlines(r)
                return r
            finally:
                if not conn.closed:
                    fcntl.fcntl(conn, fcntl.F_SETFL, flags)

message = "Other end disconnected!"

def recv_some(p, t = 5, e = 1, tr = 5, stderr = 0):
    '''
    Recieving all answers from cluster in default 5 sec.
    '''
    if tr < 1:
        tr = 1
    x = time.time() + t
    y = []
    r = ''
    pr = p.recv
    if stderr:
        pr = p.recv_err
    while time.time() < x or r:
        r = pr()
        if r is None:
            if e:
                raise Exception(message)
            else:
                break
        elif r:
            y.append(r)
        else:
            time.sleep(max((x - time.time()) / tr, 0))
    return ''.join(y)

def send_all(p, data):
    '''
    Send a row of commands to the shell.
    '''
    while len(data):
        sent = p.send(data)
        if sent is None:
            raise Exception(message)
        data = buffer(data, sent)

def delete_old(p, filename, tail):
    '''
    Delete all old files with the filename and these endings laying on the 
    cluster.
    
    p [Popen]: A active shell.
    filename [string]: Actual file.
    tail [string]: Closing escape sequence (WIN and Unix are different)
    '''
    endings = ['.com', '.log', '.cae', '.dat',
               '.jnl', '.msg', '.odb', '.prt',
               '.sim', '.sta', '.xml', '.stt',
               '.odb_f', '.lck', '.mdl', '.ipm']
    remove = 'rm '
    command_list = []
    for end in endings:
        command_list.append(remove + filename + end + tail)
    for cmd in command_list:
        send_all(p, cmd)

def solve_abaqus(p, filename, tail):
    '''
    Start a abaqus job with the *.inp file.
    
    p [Popen]: A active shell.
    filename [string]: Actual file.
    tail [string]: Closing escape sequence (WIN and Unix are different)
    '''
    send_all(p, 'abaqus job=' + filename + tail)

def open_abaqus(p, tail):
    '''
    Starts the Abaqus CAE.
    
    p [Popen]: A active shell.
    tail [string]: Closing escape sequence (WIN and Unix are different)
    '''
    send_all(p, 'abaqus cae ' + tail)

def connect_cluster(p, login, tail, cluster = 'cluster.rz.rwth-aacheb.de', options = []):
    '''
    Connects to the cluster.
    
    p [Popen]: A active shell.
    login [string]: Working login name.
    tail [string]: Closing escape sequence (WIN and Unix are different)
    cluster [string]: Path of the cluster.
    options [List]: List of option strings.
    '''
    cmd = 'ssh '
    for i in options:
        cmd += i + ' '
    cmd += cluster + ' -l ' + login + tail
    send_all(p, cmd)
    

def upload_file(p, login, path_file, tail, cluster = 'cluster.rz.rwth-aachen.de', path_server = '~/'):
    '''
    Uploads an explicit *.inp file to the cluster.
    
    p [Popen]: A active shell.
    login [string]: Working login name.
    path_file [string]: path of the file, which should be uploaded.
    tail [string]: Closing escape sequence (WIN and Unix are different)
    cluster [string]: Path of the cluster.
    path_server [string]: Path where the file should be uploaded to. Default 
                          is home.
    '''
    cmd = 'scp ' + path_file + ' ' + login + '@' + cluster + ':' + path_server + tail
    send_all(p, cmd)

def download_file(p, login, path_file, tail, path_server, cluster = 'cluster.rz.rwth-aachen.de'):
    '''
    Download an explicit *.inp file from the cluster.
    
    p [Popen]: A active shell.
    login [string]: Working login name.
    path_file [string]: path of the file, where it should be downloaded.
    tail [string]: Closing escape sequence (WIN and Unix are different)
    path_server [string]: Path where the file lays, which should be downloaded to.
    cluster [string]: Path of the cluster.
    '''
    cmd = 'scp ' + login + '@' + cluster + ':' + path_server + ' ' + path_file + tail
    send_all(p, cmd)
    
def close_connection(p, tail):
    '''
    Kills the shell.
    '''
    send_all(p, 'exit' + tail)
    p.kill()
    
def open_shell():
    '''
    Opens a new shell.
    
    Returns a Popen object with the active shell.
    '''
    if sys.platform == 'win32':
        shell, commands, tail = ('cmd', ('dir /w', 'echo HELLO WORLD'), '\r\n')
    else:
        shell, commands, tail = ('sh', ['ssh -Y -t -t cluster-x.rz.rwth-aachen.de -l ms299282', 'echo HELLO WORLD'], '\n')
    a = Popen(shell, stdin = PIPE, stdout = PIPE)
    return a


if __name__ == '__main__':
    tail = '\n'
    #open commandshell 
    a = open_shell()
    

    cluster = 'cluster-x.rz.rwth-aachen.de'
    login = 'ms299282'



    #connect to server
    options = ['-Y',
               '-t',
               '-t']

    connect_cluster(a, login, tail, cluster = cluster, options = options)

    print(recv_some(a))

    #remove old datas
    
    #solve new file
    solve_abaqus(a, 'test_name', tail)
    print(recv_some(a))

    # close connection
    send_all(a, 'exit' + tail)
    print(recv_some(a, e = 0))
    a.wait()




import traits.api as tr
import bmcs_utils.api as bu
from bmcs_shell.folding.analysis.fem.fe_triangular_mesh import FETriangularMesh
from bmcs_shell.folding.geometry.wb_shell_geometry import WBShellGeometry4P
import pygmsh
import numpy as np
import k3d
import gmsh

class WBShellFETriangularMesh(FETriangularMesh):
    """Directly mapped mesh with one-to-one mapping
    """
    name = 'WBShellFETriangularMesh'

    plot_backend = 'k3d'

    geo = bu.Instance(WBShellGeometry4P)

    I_CDij = tr.DelegatesTo('geo')
    unique_node_map = tr.DelegatesTo('geo')
    n_phi_plus = tr.DelegatesTo('geo')

    direct_mesh = bu.Bool(False, DSC=True)

    subdivision = bu.Float(3, DSC=True)

    # Will be used in the parent class. Should be here to catch GEO dependency
    show_wireframe = bu.Bool(True, GEO=True)

    ipw_view = bu.View(
        *FETriangularMesh.ipw_view.content,
        bu.Item('subdivision'),
        bu.Item('direct_mesh'),
        bu.Item('export_vtk'),
        bu.Item('show_wireframe'),
    )

    mesh = tr.Property(depends_on='state_changed')

    @tr.cached_property
    def _get_mesh(self):

        X_Id = self.geo.X_Ia
        I_Fi = self.geo.I_Fi
        mesh_size = np.linalg.norm(X_Id[1] - X_Id[0]) / self.subdivision

        with pygmsh.geo.Geometry() as geom:
            xpoints = np.array([
                geom.add_point(X_d, mesh_size=mesh_size) for X_d in X_Id
            ])
            for I_i in I_Fi:
                # Create points.

                Facet(geom, xpoints[I_i])
            #                geom.add_polygon(X_id, mesh_size=mesh_size)
            gmsh.model.geo.remove_all_duplicates()
            mesh = geom.generate_mesh()
        return mesh

    X_Id = tr.Property

    def _get_X_Id(self):
        if self.direct_mesh:
            return self.geo.X_Ia
        return np.array(self.mesh.points, dtype=np.float_)

    I_Fi = tr.Property

    def _get_I_Fi(self):
        if self.direct_mesh:
            return self.geo.I_Fi
        return self.mesh.cells[1][1]

    bc_fixed_nodes = tr.Array(np.int_, value=[])
    bc_loaded_nodes = tr.Array(np.int_, value=[])

    export_vtk = bu.Button

    @tr.observe('export_vtk')
    def write(self, event=None):
        self.mesh.write("test_shell_mesh.vtk")

    def setup_plot(self, pb):
        super(WBShellFETriangularMesh, self).setup_plot(pb)

        X_Id = self.X_Id.astype(np.float32)

        fixed_nodes = self.bc_fixed_nodes
        loaded_nodes = self.bc_loaded_nodes

        X_Ma = X_Id[fixed_nodes]
        k3d_fixed_nodes = k3d.points(X_Ma, color=0x22ffff, point_size=100)
        pb.plot_fig += k3d_fixed_nodes
        pb.objects['fixed_nodes'] = k3d_fixed_nodes

        X_Ma = X_Id[loaded_nodes]
        k3d_loaded_nodes = k3d.points(X_Ma, color=0xff22ff, point_size=100)
        pb.plot_fig += k3d_loaded_nodes
        pb.objects['loaded_nodes'] = k3d_loaded_nodes

    def update_plot(self, pb):
        super(WBShellFETriangularMesh, self).update_plot(pb)

        fixed_nodes = self.bc_fixed_nodes
        loaded_nodes = self.bc_loaded_nodes
        X_Id = self.X_Id.astype(np.float32)
        pb.objects['fixed_nodes'].positions = X_Id[fixed_nodes]
        pb.objects['loaded_nodes'].positions = X_Id[loaded_nodes]


class Facet:
    dim = 2

    def __init__(self, host, xpoints):
        # Create lines
        self.curves = [
                          host.add_line(xpoints[k], xpoints[k + 1])
                          for k in range(len(xpoints) - 1)
                      ] + [host.add_line(xpoints[-1], xpoints[0])]

        self.lines = self.curves

        self.curve_loop = host.add_curve_loop(self.curves)
        # self.surface = host.add_plane_surface(ll, holes) if make_surface else None
        self.surface = host.add_plane_surface(self.curve_loop)
        self.dim_tag = self.surface.dim_tag
        self.dim_tags = self.surface.dim_tags
        self._id = self.surface._id

    def __repr__(self):
        return "<pygmsh Polygon object>"


import k3d
import numpy as np

from bmcs_shell.folding.analysis.abaqus.abaqus_link_simple import AbaqusLink
from bmcs_shell.folding.analysis.fem.tri_xdomain_fe_mitc import TriXDomainMITC
from bmcs_shell.folding.analysis.fem.vmats2D_elastic import MATS2DElastic
from ibvpy.sim.tstep_bc import TStepBC
import bmcs_utils.api as bu
import traits.api as tr
from ibvpy.bcond import BCDof
from bmcs_shell.folding.analysis.fem.tri_xdomain_fe import TriXDomainFE
from bmcs_shell.folding.analysis.fem.bcs import BoundaryConditions
from bmcs_shell.folding.analysis.fem.vmats_shell_elastic import MATSShellElastic
from bmcs_shell.folding.analysis.fets2d_mitc import FETS2DMITC
from bmcs_shell.folding.geometry.wb_shell_geometry import WBShellGeometry4P
from bmcs_shell.folding.analysis.wb_fe_triangular_mesh import WBShellFETriangularMesh

itags_str = '+GEO,+MAT,+BC'


class WBShellAnalysis(TStepBC, bu.InteractiveModel):
    name = 'WBShellAnalysis'
    plot_backend = 'k3d'
    id = bu.Str
    """ if you saved boundary conditions for your current analysis, this id will make sure these bcs
     are loaded automatically next time you create an instance with the same id """

    h = bu.Float(10, GEO=True)
    show_wireframe = bu.Bool(True, GEO=True)

    ipw_view = bu.View(
        bu.Item('h',
                editor=bu.FloatRangeEditor(low=1, high=100, n_steps=100),
                continuous_update=False),
        bu.Item('show_wireframe'),
        time_editor=bu.ProgressEditor(run_method='run',
                                      reset_method='reset',
                                      interrupt_var='interrupt',
                                      time_var='t',
                                      time_max='t_max'),
    )

    n_phi_plus = tr.Property()

    def _get_n_phi_plus(self):
        return self.xdomain.mesh.n_phi_plus

    tree = ['geo', 'bcs', 'tmodel', 'xdomain']

    geo = bu.Instance(WBShellGeometry4P, ())

    tmodel = bu.Instance(MATS2DElastic, ())
    # tmodel = bu.Instance(MATSShellElastic, ())

    bcs = bu.Instance(BoundaryConditions)
    def _bcs_default(self):
        return BoundaryConditions(geo=self.geo, n_nodal_dofs=self.xdomain.fets.n_nodal_dofs, id=self.id)

    xdomain = tr.Property(tr.Instance(TriXDomainFE),
                          depends_on="state_changed")
    '''Discretization object.'''

    @tr.cached_property
    def _get_xdomain(self):
        # prepare the mesh generator
        # mesh = WBShellFETriangularMesh(geo=self.geo, direct_mesh=False, subdivision=2)
        mesh = WBShellFETriangularMesh(geo=self.geo, direct_mesh=True)
        # construct the domain with the kinematic strain mapper and stress integrator
        return TriXDomainFE(
            mesh=mesh,
            integ_factor=self.h,
        )

        # mesh = WBShellFETriangularMesh(geo=self.geo, direct_mesh=True)
        # mesh.fets = FETS2DMITC(a= self.h)
        # return TriXDomainMITC(
        #     mesh=mesh
        # )

    domains = tr.Property(depends_on="state_changed")

    @tr.cached_property
    def _get_domains(self):
        return [(self.xdomain, self.tmodel)]

    def reset(self):
        self.sim.reset()

    t = tr.Property()

    def _get_t(self):
        return self.sim.t

    def _set_t(self, value):
        self.sim.t = value

    t_max = tr.Property()

    def _get_t_max(self):
        return self.sim.t_max

    def _set_t_max(self, value):
        self.sim.t_max = value

    interrupt = tr.Property()

    def _get_interrupt(self):
        return self.sim.interrupt

    def _set_interrupt(self, value):
        self.sim.interrupt = value

    bc = tr.Property(depends_on="state_changed")
    # @tr.cached_property
    def _get_bc(self):
        bc_fixed, _, _ = self.bcs.bc_fixed
        bc_loaded, _, _ = self.bcs.bc_loaded
        return bc_fixed + bc_loaded

    def run(self):
        s = self.sim
        s.tloop.k_max = 10
        s.tline.step = 1
        s.tloop.verbose = False
        s.run()

    def get_max_vals(self):
        self.run()
        U_1 = self.hist.U_t[-1]
        U_max = np.max(np.fabs(U_1))
        return U_max

    def export_abaqus(self):
        al = AbaqusLink(shell_analysis=self)
        al.model_name = 'test_name'
        al.build_inp()

    def setup_plot(self, pb):
        print('analysis: setup_plot')
        X_Id = self.xdomain.mesh.X_Id
        if len(self.hist.U_t) == 0:
            U_1 = np.zeros_like(X_Id)
            print('analysis: U_I', )
        else:
            U_1 = self.hist.U_t[-1]
            U_1 = U_1.reshape(-1, self.xdomain.fets.n_nodal_dofs)[:, :3]

        X1_Id = X_Id + U_1
        X1_Id = X1_Id.astype(np.float32)

        I_Ei = self.xdomain.I_Ei.astype(np.uint32)

        # Original state mesh
        wb_mesh_0 = k3d.mesh(self.xdomain.X_Id.astype(np.float32),
                             I_Ei,
                             color=0x999999, opacity=0.5,
                             side='double')
        pb.plot_fig += wb_mesh_0
        pb.objects['wb_mesh_0'] = wb_mesh_0

        # Deformed state mesh
        wb_mesh_1 = k3d.mesh(X1_Id,
                             I_Ei,
                             color_map=k3d.colormaps.basic_color_maps.Jet,
                             attribute=U_1[:, 2],
                             color_range=[np.min(U_1), np.max(U_1)],
                             side='double')
        pb.plot_fig += wb_mesh_1
        pb.objects['wb_mesh_1'] = wb_mesh_1

        if self.show_wireframe:
            k3d_mesh_wireframe = k3d.mesh(X1_Id,
                                          I_Ei,
                                          color=0x000000,
                                          wireframe=True)
            pb.plot_fig += k3d_mesh_wireframe
            pb.objects['mesh_wireframe'] = k3d_mesh_wireframe

    def update_plot(self, pb):
        X_Id = self.xdomain.mesh.X_Id
        print('analysis: update_plot')
        if len(self.hist.U_t) == 0:
            U_1 = np.zeros_like(X_Id)
            print('analysis: U_I', )
        else:
            U_1 = self.hist.U_t[-1]
            U_1 = U_1.reshape(-1, self.xdomain.fets.n_nodal_dofs)[:, :3]

        X1_Id = X_Id + U_1
        X1_Id = X1_Id.astype(np.float32)

        I_Ei = self.xdomain.I_Ei.astype(np.uint32)

        mesh = pb.objects['wb_mesh_1']
        mesh.vertices = X1_Id
        mesh.indices = I_Ei
        mesh.attribute = U_1[:, 2]
        mesh.color_range = [np.min(U_1), np.max(U_1)]
        if self.show_wireframe:
            wireframe = pb.objects['mesh_wireframe']
            wireframe.vertices = X1_Id
            wireframe.indices = I_Ei

    def get_Pw(self):
        import numpy as np
        F_to = self.hist.F_t
        U_to = self.hist.U_t
        _, _, loaded_dofs = self.bcs.bc_loaded
        F_loaded = np.sum(F_to[:, loaded_dofs], axis=-1)
        U_loaded = np.average(U_to[:, loaded_dofs], axis=-1)
        return U_loaded, F_loaded



"""Indices: I: node index, a: coordinate, F: facet index, i: node indices of the facet"""

import random

import bmcs_utils.api as bu
import k3d
import numpy as np
import traits.api as tr

from bmcs_shell.folding.geometry.math_utils import get_rot_matrix_around_vector, get_best_rot_and_trans_3d
from bmcs_shell.folding.geometry.wb_cell.wb_cell_4p import WBCell4Param
from bmcs_shell.folding.geometry.wb_cell.wb_cell_5p import WBCell5Param
from bmcs_shell.folding.geometry.wb_cell.wb_cell_5p_v2 import \
    WBCell5ParamV2
from bmcs_shell.folding.geometry.wb_cell.wb_cell_5p_v3 import WBCell5ParamV3


class WBTessellationBase(bu.Model):
    name = 'WB Tessellation Base'

    plot_backend = 'k3d'

    # show_wireframe = bu.Bool(True, GEO=True)
    show_node_labels = bu.Bool(False, GEO=True)
    wb_cell = bu.EitherType(options=[('WBCell4Param', WBCell4Param),
                                     ('WBCell5Param', WBCell5Param),
                                     ('WBCell5ParamV2', WBCell5ParamV2),
                                     ('WBCell5ParamV3', WBCell5ParamV3)
                                     ], GEO=True)
    X_Ia = tr.DelegatesTo('wb_cell_')
    I_Fi = tr.DelegatesTo('wb_cell_')
    tree = ['wb_cell']

    event_geo = bu.Bool(True, GEO=True)

    # Note: Update traits to 6.3.2 in order for the following command to work!!
    @tr.observe('wb_cell_.+GEO', post_init=True)
    def update_after_wb_cell_GEO_changes(self, event):
        self.event_geo = not self.event_geo
        self.update_plot(self.pb)

    ipw_view = bu.View(
        bu.Item('wb_cell'),
        # bu.Item('show_wireframe'),
        bu.Item('show_node_labels'),
    )

    def _get_br_X_Ia(self, X_Ia, rot=None):
        br_X_Ia = self._get_cell_matching_v1_to_v2(X_Ia, np.array([4, 6]), np.array([5, 1]))
        return self.rotate_cell(br_X_Ia, np.array([4, 6]), self.sol[0] if rot is None else rot)

    def _get_ur_X_Ia(self, X_Ia, rot=None):
        ur_X_Ia = self._get_cell_matching_v1_to_v2(X_Ia, np.array([6, 2]), np.array([3, 5]))
        return self.rotate_cell(ur_X_Ia, np.array([6, 2]), self.sol[1] if rot is None else rot)

    def _get_ul_X_Ia(self, X_Ia, rot=None):
        br_X_Ia = self._get_cell_matching_v1_to_v2(X_Ia, np.array([5, 1]), np.array([4, 6]))
        return self.rotate_cell(br_X_Ia, np.array([5, 1]), -self.sol[0] if rot is None else rot)

    def _get_bl_X_Ia(self, X_Ia, rot=None):
        br_X_Ia = self._get_cell_matching_v1_to_v2(X_Ia, np.array([3, 5]), np.array([6, 2]))
        return self.rotate_cell(br_X_Ia, np.array([3, 5]), -self.sol[1] if rot is None else rot)

    def _get_cell_matching_v1_to_v2(self, X_Ia, v1_ids, v2_ids):
        v1_2a = np.array([X_Ia[v1_ids[0]], X_Ia[v1_ids[1]], X_Ia[0]]).T
        v2_2a = np.array([X_Ia[v2_ids[0]], X_Ia[v2_ids[1]], X_Ia[v2_ids[0]] + X_Ia[v2_ids[1]] - X_Ia[0]]).T
        rot, trans = get_best_rot_and_trans_3d(v1_2a, v2_2a)

        translated_X_Ia = trans.flatten() + np.einsum('ba, Ia -> Ib', rot, X_Ia)

        return self.rotate_cell(translated_X_Ia, v1_ids, angle=np.pi)

    def rotate_cell(self, cell_X_Ia, v1_ids, angle=np.pi):
        # Rotating around vector #######
        # 1. Bringing back to origin (because rotating is around a vector originating from origin)
        cell_X_Ia_copy = np.copy(cell_X_Ia)
        cell_X_Ia = cell_X_Ia_copy - cell_X_Ia_copy[v1_ids[1]]

        # 2. Rotating
        rot_around_v1 = get_rot_matrix_around_vector(cell_X_Ia[v1_ids[0]] - cell_X_Ia[v1_ids[1]], angle)
        cell_X_Ia = np.einsum('ba, Ia -> Ib', rot_around_v1, cell_X_Ia)

        # 3. Bringing back in position
        return cell_X_Ia + cell_X_Ia_copy[v1_ids[1]]

    sol = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_sol(self):
        # No solution is provided in base class, a default value is provided for visualization
        return np.array([np.pi, np.pi])

    # Plotting ##########################################################################

    def setup_plot(self, pb):
        self.pb = pb
        pb.clear_fig()
        I_Fi = self.I_Fi
        X_Ia = self.X_Ia
        br_X_Ia = self._get_br_X_Ia(X_Ia)
        ur_X_Ia = self._get_ur_X_Ia(X_Ia)

        self.add_cell_to_pb(pb, X_Ia, I_Fi, 'X_Ia')
        self.add_cell_to_pb(pb, br_X_Ia, I_Fi, 'br_X_Ia')
        self.add_cell_to_pb(pb, ur_X_Ia, I_Fi, 'ur_X_Ia')

    k3d_mesh = {}
    k3d_wireframe = {}
    k3d_labels = {}

    def update_plot(self, pb):
        if self.k3d_mesh:
            X_Ia = self.X_Ia.astype(np.float32)
            br_X_Ia = self._get_br_X_Ia(self.X_Ia).astype(np.float32)
            ur_X_Ia = self._get_ur_X_Ia(self.X_Ia).astype(np.float32)
            self.k3d_mesh['X_Ia'].vertices = X_Ia
            self.k3d_mesh['br_X_Ia'].vertices = br_X_Ia
            self.k3d_mesh['ur_X_Ia'].vertices = ur_X_Ia
            self.k3d_wireframe['X_Ia'].vertices = X_Ia
            self.k3d_wireframe['br_X_Ia'].vertices = br_X_Ia
            self.k3d_wireframe['ur_X_Ia'].vertices = ur_X_Ia
        else:
            self.setup_plot(pb)

    def add_cell_to_pb(self, pb, X_Ia, I_Fi, obj_name):
        plot = pb.plot_fig

        wb_mesh = k3d.mesh(X_Ia.astype(np.float32),
                           I_Fi.astype(np.uint32),
                           # opacity=0.9,
                           color=0x999999,
                           side='double')
        rand_color = random.randint(0, 0xFFFFFF)
        plot += wb_mesh

        self.k3d_mesh[obj_name] = wb_mesh

        # wb_points = k3d.points(X_Ia.astype(np.float32),
        #                          color=0x999999,
        #                        point_size=100)
        # plot +=wb_points

        if self.show_node_labels:
            texts = []
            for I, X_a in enumerate(X_Ia):
                k3d_text = k3d.text('%g' % I, tuple(X_a), label_box=False, size=0.8, color=rand_color)
                plot += k3d_text
                texts.append(k3d_text)
            self.k3d_labels[obj_name] = texts

        wb_mesh_wireframe = k3d.mesh(X_Ia.astype(np.float32),
                                     I_Fi.astype(np.uint32),
                                     color=0x000000,
                                     wireframe=True)
        plot += wb_mesh_wireframe
        self.k3d_wireframe[obj_name] = wb_mesh_wireframe



from traits.api import HasTraits, Property, cached_property, \
    Array, Int, \
    Str, Float, Dict, WeakRef

import bmcs_shell.folding.analysis.abaqus.abaqus_shell_manager as asm
import numpy as np

class AbaqusLink(HasTraits):
    # data source
    shell_analysis = WeakRef

    nodes = Property

    def _get_nodes(self):
        # return np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])
        return self.shell_analysis.xdomain.mesh.X_Id

    facets = Property

    def _get_facets(self):
        # return np.array([[0, 1, 2]])
        return self.shell_analysis.xdomain.mesh.I_Fi

    bc_fixed = Property
    def _get_bc_fixed(self):
        """
        :return: [[node_idx, bc_x, bc_y, bc_z...]]
        """
        return self.shell_analysis.bcs.bc_fixed_array

    bc_loaded = Property
    def _get_bc_loaded(self):
        """
        :return: [[node_idx, f_x, f_y, f_z...]]
        """
        return self.shell_analysis.bcs.bc_loaded_array

    thickness = Property
    def _get_thickness(self):
        return self.shell_analysis.h

    nu = Property
    def _get_nu(self):
        return self.shell_analysis.tmodel.nu

    E = Property
    def _get_E(self):
        return self.shell_analysis.tmodel.E

    # -------------------------------------------------------------------------
    # Data for model building
    # -------------------------------------------------------------------------
    element_type = Str('S3R')
    model_name = Str('Model-1')
    material_name = Str('concrete')

    materials = Property()

    @cached_property
    def _get_materials(self):
        return {"concrete": [self.E, self.nu, 2500.0e-12],  # [Young's modulus in N/mm^2, poissions ratio, density in tonne/mm3 (SI-mm units)]
                      "steel": [2.1e5, 0.3, 7880.0e-12]}

    _inp_head = Property(Str, depends_on='model_name')

    @cached_property
    def _get__inp_head(self):
        """
        Head of the input file.
        """

        head = "*Heading\n\
** Job name: Job-1 Model name: " + self.model_name + "\n\
** Generated by: AbaqusLink\n\
*Preprint, echo=NO, model=Yes, history=NO, contact=NO\n\
**\n\
** Model definition\n\
**\n"
        return head

    _inp_nodes = Property(Str, depends_on='nodes')

    @cached_property
    def _get__inp_nodes(self):
        """
        Nodelist of the input file.
        """
        n = self.nodes
        nodes = "*Node,\t nset=node-1\n"
        for i in range(len(n)):
            temp_node = ' %i,\t %.4f,\t %.4f,\t %.4f\n' % (
                i + 1, n[i][0], n[i][1], n[i][2])
            nodes += temp_node
        return nodes

    _inp_elements = Property(Str, depends_on='facets, element_type')

    @cached_property
    def _get__inp_elements(self):
        """
        Elementlist of the input file.
        """
        f = self.facets
        facets = "*Element,\t elset=STRUC,\t type=" + self.element_type + "\n"
        for i in range(len(f)):
            temp_facet = ' %i,\t %i,\t %i,\t %i\t \n' % (
                i + 1, f[i][0] + 1, f[i][1] + 1, f[i][2] + 1)
            facets += temp_facet
        return facets

    _inp_sets = Property(Str, depends_on='nodes, facets, bounded_nodes')

    @cached_property
    def _get__inp_sets(self):
        """
        Sets of the input file.
        """
        set_str = '*Nset, nset=nodes, instance=Part-A, generate\n 1,\t {},\t 1\n'.format(len(self.nodes))
        set_str += '*Elset, elset=Struc, instance=Part-A, generate\n 1,\t {},\t 1\n'.format(len(self.facets))
        set_str += '*Elset, elset=_Surf-1_SNEG,  internal, instance=Part-A, generate\n 1,\t {},\t 1\n'.format(
            len(self.facets))

        bc_sets_names = self.bc_sets_names
        for i, node_bc in enumerate(self.bc_fixed):
            node_id = int(node_bc[0]) + 1
            set_str += '*Nset, nset={}, instance=Part-A\n'.format(bc_sets_names[i])
            set_str += '{}\n'.format(node_id)

        load_sets_names = self.load_sets_names
        for i, node_load in enumerate(self.bc_loaded):
            node_id = int(node_load[0]) + 1
            set_str += '*Nset, nset={}, instance=Part-A\n'.format(load_sets_names[i])
            set_str += '{}\n'.format(node_id)
        return set_str

    bc_sets_names = Property
    def _get_bc_sets_names(self):
        return ['BC-Node-{}'.format(int(node_id) + 1) for node_id in self.bc_fixed[:, 0]]

    load_sets_names = Property
    def _get_load_sets_names(self):
        return ['Load-Node-{}'.format(int(node_id) + 1) for node_id in self.bc_loaded[:, 0]]

    _inp_section = Property(Str, depends_on='thickness, material_name')

    @cached_property
    def _get__inp_section(self):
        """
        Sections of the input file.
        """
        section = '** Section: Section-1\n'
        section += '*Shell Section, elset=Struc, material='
        section += self.material_name + '\n'
        section += '%f \n' % self.thickness
        return section

    _inp_part = Property(
        Str, depends_on='_inp_nodes, _inp_elements, _inp_section')

    @cached_property
    def _get__inp_part(self):
        """
        Parts of the input file.
        """
        part = '*Part, NAME=Part-1\n'
        part += self._inp_nodes
        part += self._inp_elements
        part += self._inp_section
        part += '*End Part\n'
        return part

    _inp_instance = Property(Str)

    @cached_property
    def _get__inp_instance(self):
        """
        Instance of the input file.
        """
        instance = '*Instance, NAME=PART-A, PART=Part-1\n'
        instance += '*End Instance\n'
        return instance

    _inp_surface = Property(Str)

    @cached_property
    def _get__inp_surface(self):
        """
        Surfaces of the input file.
        """
        surf = '*Surface, type=ELEMENT, name=Surf-1\n'
        surf += '_Surf-1_SNEG, SNEG\n'
        return surf

    _inp_material = Property(Str, depends_on='material_name')

    @cached_property
    def _get__inp_material(self):
        """
        Materials of the input file.
        """
        material = '**\n'
        material += '** MATERIALS\n'
        material += '** \n'
        material += '*Material, name='
        material += self.material_name + '\n'
        material += '*Elastic\n'
        mat_values = self.materials[self.material_name]
        material += '{},\t {} \n'.format(mat_values[0], mat_values[1])
        material += '**\n *DENSITY\n {}\n'.format(mat_values[2])
        return material

    _inp_assembly = Property(Str, depends_on='nodes, facets')

    @cached_property
    def _get__inp_assembly(self):
        """
        Assembly of the input file.
        """
        assembly = '*Assembly, NAME = Assembly1\n'
        assembly += self._inp_instance
        assembly += self._inp_sets
        assembly += self._inp_surface
        assembly += '*End Assembly\n'
        return assembly

    _inp_boundary_conditions = Property(Str)

    @cached_property
    def _get__inp_boundary_conditions(self):
        bc_txt = '** BOUNDARY CONDITIONS\n'
        bc_txt += '** \n'

        bc_sets_names = self.bc_sets_names
        for i, node_bc in enumerate(self.bc_fixed):
            bc_txt += '** Name: BC-{} Type: Displacement/Rotation\n'.format(i + 1)
            bc_txt += '*Boundary\n'
            for bc_idx, bc_value in enumerate(node_bc[1:]):
                bc_id = bc_idx + 1
                if not np.isnan(bc_value):
                    bc_txt += '{}, {}, {}, {}\n'.format(bc_sets_names[i], bc_id, bc_id, bc_value)

        bc_txt += '** \n'
        return bc_txt

    _inp_loadcases = Property(Str)

    @cached_property
    def _get__inp_loadcases(self):
        """
        Loadcases of the input file.
        """
        load = '** LOADS\n'
        load += '** \n'

        # load += '** Name: Load-1   Type: Pressure\n'
        # load += '*DLOAD\n'
        # load += 'Struc, GRAV, 9810, 0.0, 0.0, -1.0\n**\n' # 9810 is Gravity in mm/s^2'

        # Concentrated forces
        load = self._add_loads(load, moment = False)

        # Moments
        load = self._add_loads(load, moment = True)

        load += '** \n'
        return load

    def _add_loads(self, load_str, moment):
        load_sets_names = self.load_sets_names
        for i, node_load in enumerate(self.bc_loaded):
            load_str += '** Name: Load-{} Type: {}\n'.format(i + 1, 'Moment' if moment else 'Concentrated force')
            load_str += '*Cload\n'
            node_load_range = node_load[4:] if moment else node_load[1:4]
            for load_idx, load_value in enumerate(node_load_range):
                load_id = load_idx + 4 if moment else load_idx + 1
                if not np.isnan(load_value):
                    load_str += '{}, {}, {}\n'.format(load_sets_names[i], load_id, load_value)
        return load_str

    _inp_output_requests = Property(Str)

    @cached_property
    def _get__inp_output_requests(self):
        """
        Output of the input file.
        """
        out = '**\n'
        out += '** OUTPUT REQUESTS\n'
        out += '** \n'
        out += '*Restart, write, frequency=0\n'
        out += '** \n'
        out += '** FIELD OUTPUT: F-Output-1\n'
        out += '** \n'
        out += '*Output, field, variable=PRESELECT\n'
        out += '**Output, field\n'
        out += '**Node Output\n'
        out += '**CF, RF, U\n'
        out += '**Element Output\n'
        out += '**ALPHA, ALPHAN, CS11, CTSHR, MISES, MISESMAX, MISESONLY, PRESSONLY, PS, S, SF, SM, SSAVG, TRIAX, TSHR, VS\n'
        out += '*Output, field, frequency=0\n'
        out += '**\n'
        out += '** HISTORY OUTPUT: H-Output-1\n'
        out += '** \n'
        out += '*Output, history, variable=PRESELECT\n'
        out += '**NODE PRINT\n'
        out += '**U,\n'
        out += '**RF,\n'
        out += '**EL PRINT\n'
        out += '**S,\n'
        return out

    _inp_step_header = Property(Str)
    @cached_property
    def _get__inp_step_header(self):
        step_header = '** ----------------------------------------------------------------\n'
        step_header += '**\n'
        step_header += '** STEP: Step-1\n'
        step_header += '** \n'
        step_header += '*Step, name=Step-1\n'
        step_header += '*Static\n'
        step_header += '0.1, 1., 1e-05, 0.1\n'
        step_header += '**\n'
        return step_header

    _inp_step_footer = Property(Str)
    @cached_property
    def _get__inp_step_footer(self):
        return '*End Step\n'

    def build_inp(self):
        fname = self.model_name + '.inp'
        inp_file = open(fname, 'w')
        inp_file.write(self._inp_head)
        inp_file.write(self._inp_part)
        inp_file.write(self._inp_assembly)
        inp_file.write(self._inp_material)
        inp_file.write(self._inp_step_header)
        inp_file.write(self._inp_boundary_conditions)
        inp_file.write(self._inp_loadcases)
        inp_file.write(self._inp_output_requests)
        inp_file.write(self._inp_step_footer)
        inp_file.close()
        print('inp file %s written' % fname)

    # =======================================================================
    # Solving of the problem in Abaqus
    # =======================================================================

    tail = '\n'

    def abaqus_solve(self):
        p = asm.open_shell()
        asm.solve_abaqus(p, self.model_name, self.tail)
        # results file is 'dat'
        # asm.download_file(
        #     p, self.login, self.model_name + '.dat', self.tail, self.model_name + '.dat')
        p.kill()

    def abaqus_cae(self):
        pass
        # asm.open_abaqus(p, self.tail)


if __name__ == '__main__':
    from bmcs_shell.folding.analysis.fem.vmats_shell_elastic import MATSShellElastic
    from bmcs_shell.folding.analysis.wb_shell_analysis import WBShellAnalysis

    tmodel = MATSShellElastic(E=28000, nu=0.2)
    n_phi_plus = 2
    n_x_plus = 1
    wba = WBShellAnalysis(h=10, tmodel=tmodel)
    data = dict(alpha=1.24, a=500, a_high=3000, b=900, b_high=3000, c=400, c_high=2000,
                n_phi_plus=n_phi_plus, n_x_plus=n_x_plus, show_nodes=False)
    wba.geo.trait_set(**data)

    al = AbaqusLink(shell_analysis=wba)
    al.model_name = 'test_name'
    al.build_inp()
    # al.abaqus_solve()
    # al.abaqus_cae()

    # import subprocess
    #
    # # cmd = ['/run/myscript', '--arg', 'value']
    # cmd = ['abaqus cae', '--arg', 'value']
    # process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    # process.wait()
    # print(process.returncode)
    #
    # # p = asm.open_shell()
    # # asm.open_abaqus(p, '\n')



#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Jan 3, 2013 by:  schmerl

from copy import copy
from traits.api import HasTraits, Property, cached_property, \
    Array, Int, \
    Str, Float, Dict, WeakRef

import bmcs_shell.folding.analysis.abaqus.abaqus_shell_manager as asm
import numpy as np


class AbaqusLink(HasTraits):
    # data source
    data = WeakRef
    # Iterationstep is initialised as last step
    iterationstep = Int(-1)
    # Number of segments each creaseline will be splitted up
    n_split = Int(2, geometry=True)

    _cp_node_index = Property(depends_on='data.nodes, data')

    @cached_property
    def _get__cp_node_index(self):
        ''' Indexlist of all Nodes in Creasepattern Plane.

        Identify all indexes of the crease pattern nodes laying in
        z = 0 plane (in t = 0) and aren't grab points or line points.
        '''
        # @todo: modify this, the proper criterion is to test if a node
        #   is attached to a facet
        #
        return self.data.cp.N
#        index = np.ma.array(range(0, len(n)))
#
#        gp = np.array(self.data.GP)
#        if(len(gp > 0)):
#            index[gp[:, 0]] = np.ma.masked
#        lp = np.array(self.data.LP)
#        if(len(lp > 0)):
#            index[lp[:, 0]] = np.ma.masked
#        index = index.compressed()
#        final_index = []
#        for i in index:
#            if(n[i, 2] == 0):
#                final_index.append(i)
#        return final_index

    _origin_nodes = Property(Array(value=[], dtype=float, geometry=True),
                             depends_on='data.N, data, iterationstep')

    @cached_property
    def _get__origin_nodes(self):
        '''
        The origin nodes of the structure, before mesh refinement.
        '''
        return self.data.x_t[self.iterationstep][self._cp_node_index]

    _origin_facets = Property(Array(value=[], dtype='int_', geometry=True),
                              depends_on='data.F, data')

    @cached_property
    def _get__origin_facets(self):
        '''
        The origin facets of the structure, before mesh refinement.
        The facets must be aligned, so all normals shows to the same side.
        '''
        return self.data.cp.aligned_facets

    _origin_cl = Property(Array(value=[], dtype='int_', geometry=True),
                          depends_on='data.L, data')

    @cached_property
    def _get__origin_cl(self):
        '''
        For meshrefinement only creaselines in the creasepatten are necessary!
        They are laing on z=0 in t=0, so crane sticks etc. are rejected.
        '''
        cl = self.data.L
        o_n = self._cp_node_index
        final_cl = np.zeros((0, 2), dtype=int)
        for i in cl:
            test = np.in1d(o_n, i)
            # 2 means both nodes of the cl are in the index table
            if(np.sum(test) == 2):
                final_cl = np.vstack((final_cl, i))
        return final_cl

    nodes = Property

    def _get_nodes(self):
        '''
        Returns all nodes after mesh refinement
        '''
        # return self._geometry[0]
        return np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])

    facets = Property

    def _get_facets(self):
        '''
        Returns all facets after mesh refinement
        '''
        # return self._geometry[1]
        return np.array([[0, 1, 2]])

    _n_origin_nodes = Property(depends_on='_origin_nodes')

    @cached_property
    def _get__n_origin_nodes(self):
        '''
        Number of the origin nodes.
        '''
        return len(self._origin_nodes)

    _n_inner_nodes = Property(
        depends_on='_n_inner_nodes_pattern, _origin_facets')

    @cached_property
    def _get__n_inner_nodes(self):
        '''
        Number of all nodes laying in an origin facet, after mesh refinement.
        '''
        nodes = self._n_inner_nodes_pattern * len(self._origin_facets)
        return nodes

    _n_inner_nodes_pattern = Property(depends_on='n_split')

    @cached_property
    def _get__n_inner_nodes_pattern(self):
        '''
        Number of the nodes laying in one origin facet, after mesh refinement.
        '''
        n = self.n_split - 2
        n_inner_nodes = 0
        while(n > 0):
            n_inner_nodes += n
            n -= 1
        return n_inner_nodes

    _n_cl_nodes = Property(depends_on='n_split')

    @cached_property
    def _get__n_cl_nodes(self):
        '''
        Number of nodes, every creaseline will be added.
        '''
        return self.n_split - 1

    _L_pattern = Property(depends_on='n_split')

    @cached_property
    def _get__L_pattern(self):
        '''
        Pattern of the triangular coordinates of every node in origin facet.
        '''
        L = 1 / float(self.n_split)
        L2 = copy(L)
        L3 = copy(L)
        L_pattern = []
        for i in range(1, self.n_split - 1):
            for j in range(1, self.n_split - i):
                L1 = 1 - (L2 * j + L3 * i)
                L_pattern.append([[L1],
                                  [L2 * j],
                                  [L3 * i]])
        return L_pattern

    _inner_f_pattern = Property(depends_on='_L_pattern')

    @cached_property
    def _get__inner_f_pattern(self):
        '''
        Pattern of the inner facets in one origin facet.
        '''
        inner_facets_pattern = []
        l_line = self.n_split - 3
        index = 0
        line_end = l_line
        n_inner_nodes = 0
        for i in range(self.n_split - 1):
            n_inner_nodes += i
        while(index < n_inner_nodes):
            if(index == line_end):
                line_end += l_line
                l_line -= 1
            else:
                temp_facet = [index, index + 1, index + l_line + 1]
                inner_facets_pattern.append(temp_facet)
                if(index + 1 < line_end):
                    temp_facet = [
                        index + 1, index + l_line + 2, index + l_line + 1]
                    inner_facets_pattern.append(temp_facet)
            index += 1
        inner_facets_pattern = np.array(inner_facets_pattern).reshape((-1, 3))
        return inner_facets_pattern

    _geometry = Property(depends_on='+geometry')

    @cached_property
    def _get__geometry(self):
        '''
        Builds all new nodes and facets for mesh refinement.
        '''
        # distance of all new nodes in the facet
        facets = np.zeros((0, 3), dtype=Int)
        nodes = self._origin_nodes

        # build inner nodes
        for f in self._origin_facets:
            f_nodes = nodes[f]
            for L_node in self._L_pattern:
                temp_node = (f_nodes * L_node).sum(axis=0)
                nodes = np.vstack((nodes, temp_node))
        # build creaseline nodes
        for cl in self._origin_cl:
            n1, n2 = nodes[cl]
            c_vec = n2 - n1
            for i in range(1, self.n_split):
                temp = n1 + i / float(self.n_split) * c_vec
                nodes = np.vstack((nodes, temp))

        # build inner_facets
        for i in range(len(self._origin_facets)):
            startnode = self._n_origin_nodes + i * self._n_inner_nodes_pattern
            inner_facets = self._inner_f_pattern + startnode
            facets = np.vstack((facets, inner_facets))

        # build outer_facets
        outer_counter = 0
        for f in self._origin_facets:
            cl = [[f[0], f[1]],
                  [f[1], f[2]],
                  [f[2], f[0]]]
            cl_index = [-1, -1, -1]
            cl_dir = [True, True, True]
            for c in range(3):
                try:
                    cl_index[c] = self._origin_cl.tolist().index(cl[c])
                except:
                    cl_dir[c] = False
                    cl_index[c] = self._origin_cl.tolist().index(
                        [cl[c][1], cl[c][0]])

            step = list(range(self.n_split - 2))

            if cl_dir[0] is False:
                step = list(range(self.n_split - 2, 0, -1))

            counter_f = 0
            for counter in step:
                node_index_cl = (cl_index[0] * self._n_cl_nodes + counter +
                                 self._n_origin_nodes + self._n_inner_nodes)
                node_index_f = outer_counter * self._n_inner_nodes_pattern + \
                    counter_f + self._n_origin_nodes
                dir_count = -1
                if(cl_dir[0]):
                    dir_count = +1
                temp_facet_idx = (node_index_cl,
                                  node_index_cl + dir_count, node_index_f)
                temp_facet = [temp_facet_idx]

                facets = np.vstack((facets, temp_facet))
                if(counter != step[-1]):
                    temp_facet_idx = (node_index_cl + dir_count,
                                      node_index_f + 1, node_index_f)
                    temp_facet = [temp_facet_idx]
                    facets = np.vstack((facets, temp_facet))
                if(counter_f == 0):
                    node_index_cl = (cl_index[0] * self._n_cl_nodes +
                                     step[0] + self._n_origin_nodes +
                                     self._n_inner_nodes)
                    pos2 = 0
                    if(cl_dir[2]):
                        pos2 = self.n_split - 2
                    node_index_cl2 = cl_index[
                        2] * (self._n_cl_nodes + pos2 + self._n_origin_nodes +
                              self._n_inner_nodes)
                    facets = np.vstack((facets,
                                        [f[0], node_index_cl, node_index_cl2]))
                    facets = np.vstack((facets,
                                        [node_index_f, node_index_cl2,
                                         node_index_cl]))
                counter_f += 1

            step = list(range(self.n_split - 2))
            if(cl_dir[1] is False):
                step = list(range(self.n_split - 2, 0, -1))
            counter_f = 0
            f_pos = self.n_split - 3
            for counter in step:
                node_index_cl = (cl_index[1] * self._n_cl_nodes + counter +
                                 self._n_origin_nodes + self._n_inner_nodes)
                if(counter_f > 0):
                    f_pos += self.n_split - 2 - counter_f
                node_index_f = outer_counter * \
                    self._n_inner_nodes_pattern + f_pos + self._n_origin_nodes
                dir_count = -1
                if(cl_dir[1]):
                    dir_count = +1
                temp_facet = [
                    node_index_cl, node_index_cl + dir_count, node_index_f]
                facets = np.vstack((facets, temp_facet))
                if(counter != step[-1]):
                    temp_facet = [node_index_cl + dir_count, node_index_f +
                                  self.n_split - 3 - counter_f, node_index_f]
                    facets = np.vstack((facets, temp_facet))
                if(counter_f == 0):
                    node_index_cl = (cl_index[1] * self._n_cl_nodes + step[0] +
                                     self._n_origin_nodes +
                                     self._n_inner_nodes)
                    pos2 = 0
                    if(cl_dir[0]):
                        pos2 = self.n_split - 2
                    node_index_cl2 = (cl_index[0] * self._n_cl_nodes + pos2 +
                                      self._n_origin_nodes +
                                      self._n_inner_nodes)
                    facets = np.vstack(
                        (facets, [f[1], node_index_cl, node_index_cl2]))
                    facets = np.vstack((facets,
                                        [node_index_f, node_index_cl2,
                                         node_index_cl]))
                counter_f += 1

            step = list(range(1, self.n_split - 1))
            if(cl_dir[2]):
                step = list(range(self.n_split - 2))
                step.reverse()

            counter_f = 0
            for counter in step:
                node_index_cl = cl_index[
                    2] * self._n_cl_nodes + counter + self._n_origin_nodes + self._n_inner_nodes
                f_pos = 0
                for i in range(counter_f):
                    f_pos += (self.n_split - 2) - i
                node_index_f = outer_counter * \
                    self._n_inner_nodes_pattern + f_pos + self._n_origin_nodes
                dir_count = -1
                if(cl_dir[2]):
                    dir_count = +1
                temp_facet = [
                    node_index_cl, node_index_cl + dir_count, node_index_f]
                facets = np.vstack((facets, temp_facet))
                if(counter != step[-1]):
                    temp_facet = [
                        node_index_cl, node_index_f, node_index_f + self.n_split - 2 - counter_f]
                    facets = np.vstack((facets, temp_facet))
                if(counter == step[-1]):
                    node_index_cl = cl_index[
                        2] * self._n_cl_nodes + step[-1] + self._n_origin_nodes + self._n_inner_nodes
                    pos2 = 0
                    if(cl_dir[1]):
                        pos2 = self.n_split - 2
                    node_index_cl2 = cl_index[
                        1] * self._n_cl_nodes + pos2 + self._n_origin_nodes + self._n_inner_nodes
                    facets = np.vstack(
                        (facets, [f[2], node_index_cl, node_index_cl2]))
                    facets = np.vstack(
                        (facets, [node_index_f, node_index_cl2, node_index_cl]))
                counter_f += 1
            outer_counter += 1
        return [nodes, facets]

    #-------------------------------------------------------------------------
    # Data for model building
    #-------------------------------------------------------------------------
    element_type = Str('S3R')
    model_name = Str('Model-1')
    material_name = Str('concrete')
    materials = Dict({"concrete": [3.0e10, 0.2, 2500],  # [Young's modulus in N/m3, poissions ratio, density in kg/m3]
                      "steel": [21.0e10, 0.3, 7880]})
    thickness = Float(0.06)  # Thickness in meter
    # bounded_nodes = Array(np.int, value=[1, 2, 3, 13, 14, 15])
    bounded_nodes = Array(np.int, value=[0, 2])

    _inp_head = Property(Str, depends_on='model_name')

    @cached_property
    def _get__inp_head(self):
        '''
        Head of the input file.
        '''

        head = "*Heading\n\
** Job name: Job-1 Model name: " + self.model_name + "\n\
** Generated by: AbaqusLink\n\
*Preprint, echo=NO, model=Yes, history=NO, contact=NO\n\
**\n\
** Model definition\n\
**\n"
        return head

    _inp_nodes = Property(Str, depends_on='nodes')

    @cached_property
    def _get__inp_nodes(self):
        '''
        Nodelist of the input file.
        '''
        n = self.nodes
        nodes = "*Node,\t nset=node-1\n"
        for i in range(len(n)):
            temp_node = ' %i,\t %.4f,\t %.4f,\t %.4f\n' % (
                i + 1, n[i][0], n[i][1], n[i][2])
            nodes += temp_node
        return nodes

    _inp_elements = Property(Str, depends_on='facets, element_type')

    @cached_property
    def _get__inp_elements(self):
        '''
        Elementlist of the input file.
        '''
        f = self.facets
        facets = "*Element,\t elset=STRUC,\t type=" + self.element_type + "\n"
        for i in range(len(f)):
            temp_facet = ' %i,\t %i,\t %i,\t %i\t \n' % (
                i + 1, f[i][0] + 1, f[i][1] + 1, f[i][2] + 1)
            facets += temp_facet
        return facets

    _inp_sets = Property(Str, depends_on='nodes, facets, bounded_nodes')

    @cached_property
    def _get__inp_sets(self):
        '''
        Sets of the input file.
        '''
        set_str = '*Nset, nset=nodes, instance=Part-A, generate\n 1,\t %i,\t 1\n\
*Elset, elset=Struc, instance=Part-A, generate\n 1,\t %i,\t 1\n\
*Elset, elset=_Surf-1_SNEG,  internal, instance=Part-A, generate\n 1,\t %i,\t 1\n' % (len(self.nodes), len(self.facets), len(self.facets))
        set_str += '*Nset, nset=boundery, instance=Part-A\n'
        for i in self.bounded_nodes:
            print(i)
            set_str += '%i, ' % (i)
        set_str += '\n'
        return set_str

    _inp_section = Property(Str, depends_on='thickness, material_name')

    @cached_property
    def _get__inp_section(self):
        '''
        Sections of the input file.
        '''
        section = "** Section: Section-1\n\
*Shell Section, elset=Struc, material="
        section += self.material_name + '\n'
        section += '%f \n' % (self.thickness)
        return section

    _inp_part = Property(
        Str, depends_on='_inp_nodes, _inp_elements, _inp_section')

    @cached_property
    def _get__inp_part(self):
        '''
        Parts of the input file.
        '''
        part = '*Part, NAME=Part-1\n'
        part += self._inp_nodes
        part += self._inp_elements
        part += self._inp_section
        part += '*end Part\n'
        return part

    _inp_instance = Property(Str)

    @cached_property
    def _get__inp_instance(self):
        '''
        Instance of the input file.
        '''
        instance = '*Instance, NAME=PART-A, PART=Part-1\n\
*END INSTANCE\n'
        return instance

    _inp_surface = Property(Str)

    @cached_property
    def _get__inp_surface(self):
        '''
        Surfaces of the input file.
        '''
        surf = '*Surface, type=ELEMENT, name=Surf-1\n\
_Surf-1_SNEG, SNEG\n'
        return surf

    _inp_material = Property(Str, depends_on='material_name')

    @cached_property
    def _get__inp_material(self):
        '''
        Materials of the input file.
        '''
        material = '**\n\
** MATERIALS\n\
** \n\
*Material, name='
        material += self.material_name + '\n'
        material += '*Elastic\n'
        mat_values = self.materials[self.material_name]
        material += '%f,\t %f \n' % (mat_values[0], mat_values[1])
        material += '**\n *DENSITY\n %f\n' % (mat_values[2])
        material += '** ----------------------------------------------------------------\n'
        return material

    _inp_assembly = Property(Str, depends_on='nodes, facets')

    @cached_property
    def _get__inp_assembly(self):
        '''
        Assembly of the input file.
        '''
        assembly = '*Assembly, NAME = Assembly1\n'
        assembly += self._inp_instance
        assembly += self._inp_sets
        assembly += self._inp_surface
        assembly += '*End Assembly\n'
        return assembly

    _inp_loadcases = Property(Str)

    @cached_property
    def _get__inp_loadcases(self):
        '''
        Loadcases of the input file.
        '''
        load = '**\n\
** STEP: Step-1\n\
** \n\
*Step, name=Step-1\n\
*Static\n\
0.1, 1., 1e-05, 0.1\n\
**\n\
** \n\
** BOUNDARY CONDITIONS\n\
** \n\
** Name: BC-1\n\
*Boundary\n\
boundery, Encastre\n\
** \n\
** LOADS\n\
** \n\
** Name: Load-1   Type: Pressure\n\
*DLOAD\n\
Struc, GRAV, 9.81, 0.0, 0.0, -1.0\n'
        return load

    _inp_output = Property(Str)

    @cached_property
    def _get__inp_output(self):
        '''
        Output of the input file.
        '''
        out = '**\n\
** OUTPUT REQUESTS\n\
** \n\
*Restart, write, frequency=0\n\
** \n\
** FIELD OUTPUT: F-Output-1\n\
** \n\
*Output, field\n\
#*Node Output\n\
#CF, RF, U\n\
#*Element Output\n\
#ALPHA, ALPHAN, CS11, CTSHR, MISES, MISESMAX, MISESONLY, PRESSONLY, PS, S, SF, SM, SSAVG, TRIAX, TSHR, VS\n\
*Output, field, frequency=0\n\
**\n\
** HISTORY OUTPUT: H-Output-1\n\
** \n\
*Output, history, variable=PRESELECT\n\
*NODE PRINT\n\
U,\n\
RF,\n\
*EL PRINT\n\
S,\n\
*End Step'
        return out

    def build_inp(self):
        # Data head
        head = self._inp_head
        # part
        part = self._inp_part
        # Assembly
        assembly = self._inp_assembly
        # Material
        material = self._inp_material
        # Data bottom
        loadcases = self._inp_loadcases
        output = self._inp_output

        fname = self.model_name + '.inp'
        inp_file = open(fname, 'w')
        inp_file.write(head)
        inp_file.write(part)
        inp_file.write(assembly)
        inp_file.write(material)
        inp_file.write(loadcases)
        inp_file.write(output)

        inp_file.close()
        print('inp file %s written' % fname)

#=======================================================================
# Connection to server and solving of the problem
#=======================================================================
    # Abaqus Shell Manager Datas:
    cluster = 'cluster-x.rz.rwth-aachen.de'
    login = 'ms299282'
    options = ['-Y',
               '-t',
               '-t']

    tail = '\n'

    def abaqus_solve(self):
        ''' Solve FE with Abaqus on RWTH Cluster.

        The previouse generated input file will be uploaded on the cluster System
        of the RWTH and solved by Abaqus. Finally the result file will be
        downloaded.

        .. note:: You need a public key connection to the Cluster System. An intsallationguide
        is following at the end. You need this key on every computer you want to run this
        code from.

        .. note:: You have to change the login ID to your own TIM ID.

        .. note:: This code can only be run from an Unix machine!

        ToDo:
        - enable this code to Windows
        - make a better output of cluster
        - implement exception handling
        - killing the connection
        '''
        # Initialise connection
        p = asm.open_shell()
        # establish connection
        asm.connect_cluster(
            p, self.login, self.tail, cluster=self.cluster, options=self.options)
        print(asm.recv_some(p))
        # delete old files with same name
        asm.delete_old(p, self.model_name, self.tail)
        print(asm.recv_some(p))
        # upload the input file
        asm.upload_file(p, self.login, self.model_name + '.inp', self.tail)
        print(asm.recv_some(p))
        # start abaqus solver on server
        asm.solve_abaqus(p, self.model_name, self.tail)
        print(asm.recv_some(p))
        # close connection
        asm.close_connection(p, self.tail)
        # open new connection for downloading results
        p = asm.open_shell()
        asm.download_file(
            p, self.login, self.model_name + '.dat', self.tail, self.model_name + '.dat')
        print(asm.recv_some(p))
        p.kill()

    def abaqus_cae(self):
        # Initialise connection
        p = asm.open_shell()
        # establish connection
        asm.connect_cluster(
            p, self.login, self.tail, cluster=self.cluster, options=self.options)
        print(asm.recv_some(p))
        asm.open_abaqus(p, self.tail)
        print(asm.recv_some(p))
        asm.close_connection(p, self.tail)


#=========================================================================
# Installation guide for a publickey for autoconnection on Unix Systems
#=========================================================================
'''
This guide is copied from:
http://wp.uberdose.com/2006/10/16/ssh-automatic-login/

SSH Automatic Login

Of course this is not the right phrase for it. It should be something like
key-based authorization with SSH. Or simply publickey authorization.
Or unattended ssh login. But I guess you know what I mean.

Here are the steps:

1.  Create a public ssh key, if you haven't one already.
    Look at ~/.ssh. If you see a file named id_dsa.pub then you obviously
    already have a public key. If not, simply create one. ssh-keygen -t dsa
    should do the trick.
    Please note that there are other types of keys, e.g. RSA instead of DSA.
    I simply recomend DSA, but keep that in mind if you run into errors.
2.  Make sure your .ssh dir is 700:
    chmod 700 ~/.ssh
3.  Get your public ssh key on the server you want to login automatically.
    A simple scp ~/.ssh/id_dsa.pub remoteuser@remoteserver.com: is ok.
4.  Append the contents of your public key to the ~/.ssh/authorized_keys and remove it.
    Important: This must be done on the server you just copied your public key to.
    Otherwise you wouldn't have had to copy it on your server.
    Simply issue something like cat id_dsa.pub >> .ssh/authorized_keys while at your home directory.
5.  Instead of steps 3 and 4, you can issue something like this:
    cat ~/.ssh/id_dsa.pub | ssh -l remoteuser remoteserver.com 'cat >> ~/.ssh/authorized_keys'
6.  Remove your public key from the home directory on the server.
7.  Done!
    You can now login:
    ssh -l remoteuser remoteserver.com or ssh remoteuser@remoteserver.com without getting asked for a password.

That's all you need to do.

'''


if __name__ == '__main__':
    from oricreate.factories import YoshimuraCPFactory
#    from oricrete.FoldRigidly2.examples.YoshimuraCreasePattern.ex03_rhombus_ref_surface import create_cp_fc_03
#    from oricrete.FoldRigidly2.FoldRigidlyphase import Lifting
    cp = YoshimuraCPFactory(L_x=4, L_y=2, n_x=4, n_y=4)
    al = AbaqusLink(data=cp, n_split=10)
    al.model_name = 'test_name'
    al.build_inp()
    al.abaqus_solve()
#    al.abaqus_cae()



"""

"""
import bmcs_utils.api as bu
import sympy as sp
from bmcs_shell.folding.geometry.wb_cell.wb_cell import WBCell
from sympy.algebras.quaternion import Quaternion
import k3d
import traits.api as tr
import numpy as np
import math

def get_x_sol(Eq_UOU, x_ul, subs_yz):
    Eq_UOU_x = Eq_UOU.subs(subs_yz)
    Eq_UOU_x_rearr = sp.Eq(-Eq_UOU_x.args[1].args[1],
                           -Eq_UOU_x.args[0] + Eq_UOU_x.args[1].args[0] + Eq_UOU_x.args[1].args[2])
    Eq_UOU_x_rhs = Eq_UOU_x_rearr.args[1] ** 2 - Eq_UOU_x_rearr.args[0] ** 2
    Eq_UOU_x_rhs_collect = sp.collect(sp.expand(Eq_UOU_x_rhs), x_ul)

    A_ = Eq_UOU_x_rhs_collect.coeff(x_ul, 2)
    B_ = Eq_UOU_x_rhs_collect.coeff(x_ul, 1)
    C_ = Eq_UOU_x_rhs_collect.coeff(x_ul, 0)
    A, B, C = sp.symbols('A, B, C')

    x_ul_sol1, x_ul_sol2 = sp.solve(A * x_ul ** 2 + B * x_ul + C, x_ul)
    return x_ul_sol1, x_ul_sol2, A_, B_, C_


class WBCellSymb5ParamXL(bu.SymbExpr):

    a, b, c = sp.symbols('a, b, c', positive=True)
    gamma = sp.symbols('gamma')

    U_ur_0 = sp.Matrix([a, b, 0])
    U_ul_0 = sp.Matrix([-a, b, 0])
    V_r_0 = sp.Matrix([c, 0, 0])
    V_l_0 = sp.Matrix([-c, 0, 0])

    x_ur, y_ur, z_ur = sp.symbols(r'x_ur, y_ur, z_ur')
    x_ul, y_ul, z_ul = sp.symbols(r'x_ul, y_ul, z_ul')

    U_ur_1 = sp.Matrix([x_ur, y_ur, z_ur])
    U_ul_1 = sp.Matrix([x_ul, y_ul, z_ul])
    V_r_1 = sp.Matrix([c * sp.sin(gamma), 0, c * sp.cos(gamma)])
    V_l_1 = sp.Matrix([-c * sp.sin(gamma), 0, c * sp.cos(gamma)])
    X_UOV_r_0 = U_ur_0.T * V_r_0
    X_VOU_l_0 = U_ul_0.T * V_l_0
    X_UOV_r_1 = U_ur_1.T * V_r_1
    X_VOU_l_1 = U_ul_1.T * V_l_1
    Eq_UOV_r = sp.Eq(X_UOV_r_0[0], X_UOV_r_1[0])
    Eq_UOV_l = sp.Eq(X_VOU_l_0[0], X_VOU_l_1[0])
    X_VUO_r_0 = (V_r_0 - U_ur_0).T * (-U_ur_0)
    X_VUO_l_0 = (V_l_0 - U_ul_0).T * (-U_ul_0)
    X_VUO_r_1 = (V_r_1 - U_ur_1).T * (-U_ur_1)
    X_VUO_l_1 = (V_l_1 - U_ul_1).T * (-U_ul_1)
    Eq_VUO_r = sp.Eq(-X_VUO_r_0[0], -X_VUO_r_1[0])
    Eq_VUO_l = sp.Eq(-X_VUO_l_0[0], -X_VUO_l_1[0])

    X_UOU_0 = (U_ul_0).T * (U_ur_0)
    X_UOU_1 = (U_ul_1).T * (U_ur_1)
    Eq_UOU = sp.Eq(X_UOU_0[0], X_UOU_1[0])

    yz_ur_sol1, yz_ur_sol2 = sp.solve({Eq_UOV_r, Eq_VUO_r}, [y_ur, z_ur])
    yz_ul_sol1, yz_ul_sol2 = sp.solve({Eq_UOV_l, Eq_VUO_l}, [y_ul, z_ul])

    y_ur_sol1, z_ur_sol = yz_ur_sol1
    y_ul_sol1, z_ul_sol = yz_ul_sol1
    y_ur_sol2, _ = yz_ur_sol2
    y_ul_sol2, _ = yz_ul_sol2

    subs_yz1 = {y_ur: y_ur_sol1, z_ur: z_ur_sol,
               y_ul: y_ul_sol1, z_ul: z_ul_sol}
    subs_yz2 = {y_ur: y_ur_sol2, z_ur: z_ur_sol,
               y_ul: y_ul_sol2, z_ul: z_ul_sol}

    A, B, C = sp.symbols('A, B, C')
    x_ul_sol11, x_ul_sol12, A1_, B1_, C1_ = get_x_sol(Eq_UOU, x_ul, subs_yz1)
    x_ul_sol21, x_ul_sol22, A2_, B2_, C2_ = get_x_sol(Eq_UOU, x_ul, subs_yz2)

    x_ul11_ = x_ul_sol11
    x_ul12_ = x_ul_sol12
    x_ul21_ = x_ul_sol21
    x_ul22_ = x_ul_sol22
    y_ur1_ = y_ur_sol1
    y_ul1_ = y_ul_sol1
    y_ur2_ = y_ur_sol2
    y_ul2_ = y_ul_sol2
    z_ur_ = z_ur_sol
    z_ul_ = z_ul_sol

    P_1 = sp.sin(gamma) * x_ur - a
    P_2 = (x_ur - a * sp.sin(gamma))**2 - sp.cos(gamma)**2 * b**2
    P_3 = sp.sin(gamma) *(a**2+b**2) * x_ur - a * (a**2-b**2)

    symb_model_params = ['gamma', 'x_ur', 'a', 'b', 'c', ]
    symb_expressions = [
        ('x_ul11_', ('A', 'B', 'C')),
        ('x_ul12_', ('A', 'B', 'C')),
        ('x_ul21_', ('A', 'B', 'C')),
        ('x_ul22_', ('A', 'B', 'C')),
        ('y_ur1_', ('x_ul',)),
        ('y_ul1_', ('x_ul',)),
        ('y_ur2_', ('x_ul',)),
        ('y_ul2_', ('x_ul',)),
        ('z_ur_', ('x_ul',)),
        ('z_ul_', ('x_ul',)),
        ('A1_', ()),
        ('B1_', ()),
        ('C1_', ()),
        ('A2_', ()),
        ('B2_', ()),
        ('C2_', ()),
        ('V_r_1', ()),
        ('V_l_1', ()),
        ('P_1',()),
        ('P_2', ()),
        ('P_3', ()),
    ]

class WBCell5Param(WBCell, bu.InjectSymbExpr):
    name = 'waterbomb cell 5p'
    symb_class = WBCellSymb5ParamXL

    plot_backend = 'k3d'

    gamma = bu.Float(1, GEO=True)
    x_ur = bu.Float(1000, GEO=True)
    a = bu.Float(1000, GEO=True)
    b = bu.Float(1000, GEO=True)
    c = bu.Float(1000, GEO=True)
    a_low = bu.Float(2000)
    b_low = bu.Float(2000)
    c_low = bu.Float(2000)
    a_high = bu.Float(2000)
    b_high = bu.Float(2000)
    c_high = bu.Float(2000)
    y_sol1 = bu.Bool(False, GEO=True)
    x_sol1 = bu.Bool(False, GEO=True)

    continuous_update = True

    ipw_view = bu.View(
        bu.Item('gamma', latex=r'\gamma', editor=bu.FloatRangeEditor(
            low=1e-6, high=np.pi / 2, n_steps=100, continuous_update=continuous_update)),
        bu.Item('x_ur', latex=r'x^\urcorner', editor=bu.FloatRangeEditor(
            low=-2000, high=3000, n_steps=100, continuous_update=continuous_update)),
        bu.Item('a', latex='a', editor=bu.FloatRangeEditor(
            low=1e-6, high_name='a_high', n_steps=100, continuous_update=continuous_update)),
        bu.Item('b', latex='b', editor=bu.FloatRangeEditor(
            low=1e-6, high_name='b_high', n_steps=100, continuous_update=continuous_update)),
        bu.Item('c', latex='c', editor=bu.FloatRangeEditor(
            low=1e-6, high_name='c_high', n_steps=100, continuous_update=continuous_update)),
        bu.Item('y_sol1'),
        bu.Item('x_sol1'),
        *WBCell.ipw_view.content,
    )

    n_I = tr.Property
    def _get_n_I(self):
        return len(self.X_Ia)

    X_Ia = tr.Property(depends_on='+GEO')
    '''Array with nodal coordinates I - node, a - dimension
    '''
    @tr.cached_property
    def _get_X_Ia(self):
        gamma = self.gamma
        alpha = np.pi/2 - gamma

        P_1 = self.symb.get_P_1()
        P_2 = self.symb.get_P_2()
        P_3 = self.symb.get_P_3()

#        print('P', P_1, P_2, P_3, P_1*P_2*P_3)

        x_ur = self.x_ur

        if self.y_sol1:
            A = self.symb.get_A1_()
            B = self.symb.get_B1_()
            C = self.symb.get_C1_()
            if self.x_sol1:
                x_ul = self.symb.get_x_ul11_(A, B, C)
            else:
                x_ul = self.symb.get_x_ul12_(A, B, C)
            y_ul = self.symb.get_y_ul1_(x_ul)
            y_ur = self.symb.get_y_ur1_(x_ul)
        else:
            A = self.symb.get_A2_()
            B = self.symb.get_B2_()
            C = self.symb.get_C2_()
            if self.x_sol1:
                x_ul = self.symb.get_x_ul21_(A, B, C)
            else:
                x_ul = self.symb.get_x_ul22_(A, B, C)
            y_ul = self.symb.get_y_ul2_(x_ul)
            y_ur = self.symb.get_y_ur2_(x_ul)

        z_ur = self.symb.get_z_ur_(x_ul)
        z_ul = self.symb.get_z_ul_(x_ul)

        x_ll = -x_ur
        x_lr = -x_ul
        y_ll = - y_ur
        y_lr = - y_ul
        z_ll = z_ur
        z_lr = z_ul

        V_r_1 = self.symb.get_V_r_1().flatten()
        V_l_1 = self.symb.get_V_l_1().flatten()

        return np.array([
            [0,0,0], # 0 point
            [x_ur, y_ur, z_ur], #U++
            [x_ul, y_ul, z_ul], #U-+  ul
            [x_lr, y_lr, z_lr], #U+-
            [x_ll, y_ll, z_ll], #U--
            V_r_1, 
            V_l_1,
            ], dtype=np.float_
        )

    I_boundary = tr.Array(np.int_, value=[[2,1],
                                          [6,5],
                                          [4,3],])
    '''Boundary nodes in 2D array to allow for generation of shell boundary nodes'''

    X_theta_Ia = tr.Property(depends_on='+GEO')
    '''Array with nodal coordinates I - node, a - dimension
    '''
    @tr.cached_property
    def _get_X_theta_Ia(self):
        D_a = self.symb.get_D_(self.alpha).T
        theta = self.symb.get_theta_sol(self.alpha)
        XD_Ia = D_a + self.X_Ia
        X_center = XD_Ia[1,:]

        rotation_axes = np.array([[1, 0, 0]], dtype=np.float_)
        rotation_angles = np.array([-theta], dtype=np.float_)
        rotation_centers = np.array([X_center], dtype=np.float_)

        x_single = np.array([XD_Ia], dtype='f')
        x_pulled_back = x_single - rotation_centers[:, np.newaxis, :]
        q = axis_angle_to_q(rotation_axes, rotation_angles)
        x_rotated = qv_mult(q, x_pulled_back)
        x_pushed_forward = x_rotated + rotation_centers[:, np.newaxis, :]
        x_translated = x_pushed_forward #  + self.translations[:, np.newaxis, :]
        return x_translated[0,...]

    delta_x = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_delta_x(self):
        return self.symb.get_delta_x()

    delta_phi = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_delta_phi(self):
        return self.symb.get_delta_phi()

    R_0 = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_R_0(self):
        return self.symb.get_R_0()

def q_normalize(q, axis=1):
    sq = np.sqrt(np.sum(q * q, axis=axis))
    sq[np.where(sq == 0)] = 1.e-19
    return q / sq[:, np.newaxis]


def v_normalize(q, axis=1):
    sq = np.einsum('...a,...a->...', q, q)
    sq[np.where(sq == 0)] = 1.e-19
    return q / sq[..., np.newaxis]


def q_mult(q1, q2):
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2
    w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
    x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
    y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
    z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
    return np.array([w, x, y, z], dtype='f')


def q_conjugate(q):
    qn = q_normalize(q.T).T
    w, x, y, z = qn
    return np.array([w, -x, -y, -z], dtype='f')


def qv_mult(q1, u):
    #print('q1', q1.shape, 'u', u.shape)
    zero_re = np.zeros((u.shape[0], u.shape[1]), dtype='f')
    #print('zero_re', zero_re.shape)
    q2 = np.concatenate([zero_re[:, :, np.newaxis], u], axis=2)
    #print('q2', q2.shape)
    q2 = np.rollaxis(q2, 2)
    #print('q2', q2.shape)
    q12 = q_mult(q1[:, :, np.newaxis], q2[:, :, :])
    #print('q12', q12.shape)
    q_con = q_conjugate(q1)
    #print('q_con', q_con.shape)
    q = q_mult(q12, q_con[:, :, np.newaxis])
    #print('q', q.shape)
    q = np.rollaxis(np.rollaxis(q, 2), 2)
    #print('q', q.shape)
    return q[:, :, 1:]


def axis_angle_to_q(v, theta):
    v_ = v_normalize(v, axis=1)
    x, y, z = v_.T
#    print('x,y,z', x, y, z)
    theta = theta / 2
#    print('theta', theta)
    w = np.cos(theta)
    x = x * np.sin(theta)
    y = y * np.sin(theta)
    z = z * np.sin(theta)
#    print('x,y,z', x, y, z)
    return np.array([w, x, y, z], dtype='f')


def q_to_axis_angle(q):
    w, v = q[0, :], q[1:, :]
    theta = np.arccos(w) * 2.0
    return theta, v_normalize(v)




from bmcs_shell.folding.analysis.wb_shell_analysis import WBShellAnalysis
from bmcs_shell.folding.analysis.wb_fe_triangular_mesh import WBShellFETriangularMesh
from bmcs_shell.folding.analysis.fem.fe_triangular_mesh import FETriangularMesh
from bmcs_shell.folding.analysis.fem.tri_xdomain_fe import TriXDomainFE
from bmcs_shell.folding.analysis.fem.tri_xdomain_fe_mitc import TriXDomainMITC
from bmcs_shell.folding.analysis.fets2d_mitc import FETS2DMITC
from bmcs_shell.folding.analysis.fem.xdomain_fe_grid import XDomainFE
from bmcs_shell.folding.analysis.wb_fets2d3u1m_fe import FETS2D3U1M
from bmcs_shell.folding.geometry.wb_cell.wb_cell import WBCell
from bmcs_shell.folding.geometry.wb_cell.wb_cell_4p import WBCell4Param
from bmcs_shell.folding.geometry.wb_cell.wb_cell_5p import WBCell5Param
from bmcs_shell.folding.geometry.wb_cell.wb_cell_5p import WBCellSymb5ParamXL
from bmcs_shell.folding.geometry.wb_cell.wb_cell_5p_v2 import WBCell5ParamV2
from bmcs_shell.folding.geometry.wb_cell.wb_cell_5p_v3 import WBCell5ParamV3
from bmcs_shell.folding.geometry.wb_tessellation.wb_tessellation_4p import WBTessellation4P
from bmcs_shell.folding.analysis.fem.vmats2D_elastic import MATS2DElastic
from bmcs_shell.folding.analysis.fem.vmats_shell_elastic import MATSShellElastic
from bmcs_shell.folding.geometry.wb_tessellation.wb_num_tessellation_base import WBNumTessellationBase
from bmcs_shell.folding.geometry.wb_tessellation.wb_tessellation_base import WBTessellationBase
from bmcs_shell.folding.geometry.wb_tessellation.wb_num_tessellation_invest import WBNumTessellationInvest
from bmcs_shell.folding.geometry.wb_tessellation.wb_num_tessellation import WBNumTessellation
from bmcs_shell.folding.geometry.wb_tessellation.wb_tessellation_5pv2 import WBTessellation5PV2
from bmcs_shell.folding.analysis.abaqus.abaqus_link_simple import AbaqusLink



import bmcs_utils.api as bu
import numpy as np

from bmcs_shell.folding.geometry.wb_tessellation.wb_num_tessellation_base import WBNumTessellationBase


class WBNumTessellationInvest(WBNumTessellationBase):
    """ A class to investigate the angles for tessellating three cells manually. """

    name = 'WBNumTessellationInvest'

    rot_br = bu.Float(0.5)
    rot_ur = bu.Float(0.5)
    investigate_rot = bu.Bool

    ipw_view = bu.View(
        *WBNumTessellationBase.ipw_view.content,
        bu.Item('investigate_rot'),
        bu.Item('rot_br', latex=r'rot~br', editor=bu.FloatRangeEditor(low=0, high=2 * np.pi, n_steps=150,
                                                                      continuous_update=True)),
        bu.Item('rot_ur', latex=r'rot~ur', editor=bu.FloatRangeEditor(low=0, high=2 * np.pi, n_steps=150,
                                                                      continuous_update=True)),
    )

    def setup_plot(self, pb):
        super().setup_plot(pb)

    def update_plot(self, pb):
        if self.k3d_mesh:
            sol = self.sol
            X_Ia = self.X_Ia.astype(np.float32)
            br_X_Ia = self._get_br_X_Ia(self.X_Ia, self.rot_br if self.investigate_rot else sol[
                0]).astype(np.float32)
            ur_X_Ia = self._get_ur_X_Ia(self.X_Ia, self.rot_ur if self.investigate_rot else sol[
                1]).astype(np.float32)
            self.k3d_mesh['X_Ia'].vertices = X_Ia
            self.k3d_mesh['br_X_Ia'].vertices = br_X_Ia
            self.k3d_mesh['ur_X_Ia'].vertices = ur_X_Ia
            self.k3d_wireframe['X_Ia'].vertices = X_Ia
            self.k3d_wireframe['br_X_Ia'].vertices = br_X_Ia
            self.k3d_wireframe['ur_X_Ia'].vertices = ur_X_Ia
        else:
            self.setup_plot(pb)
